"""
Test various resource downloading functions from resources/common.py
"""

import os
import pytest
import tempfile

import stanza
from stanza.resources import common
from stanza.tests import TEST_MODELS_DIR, TEST_WORKING_DIR

pytestmark = [pytest.mark.travis, pytest.mark.client]

def test_assert_file_exists():
    with tempfile.TemporaryDirectory(dir=TEST_WORKING_DIR) as test_dir:
        filename = os.path.join(test_dir, "test.txt")
        with pytest.raises(FileNotFoundError):
            common.assert_file_exists(filename)

        with open(filename, "w", encoding="utf-8") as fout:
            fout.write("Unban mox opal!")
        # MD5 of the fake model file, not any real model files in the system
        EXPECTED_MD5 = "44dbf21b4e89cea5184615a72a825a36"
        common.assert_file_exists(filename)
        common.assert_file_exists(filename, md5=EXPECTED_MD5)

        with pytest.raises(ValueError):
            common.assert_file_exists(filename, md5="12345")

        with pytest.raises(ValueError):
            common.assert_file_exists(filename, md5="12345", alternate_md5="12345")

        common.assert_file_exists(filename, md5="12345", alternate_md5=EXPECTED_MD5)


def test_download_tokenize_mwt():
    with tempfile.TemporaryDirectory(dir=TEST_WORKING_DIR) as test_dir:
        stanza.download("en", model_dir=test_dir, processors="tokenize", package="ewt", verbose=False)
        pipeline = stanza.Pipeline("en", model_dir=test_dir, processors="tokenize", package="ewt")
        assert isinstance(pipeline, stanza.Pipeline)
        # mwt should be added to the list
        assert len(pipeline.loaded_processors) == 2

def test_download_non_default():
    """
    Test the download path for a single file rather than the default zip

    The expectation is that an NER model will also download two charlm models.
    If that layout changes on purpose, this test will fail and will need to be updated
    """
    with tempfile.TemporaryDirectory(dir=TEST_WORKING_DIR) as test_dir:
        stanza.download("en", model_dir=test_dir, processors="ner", package="ontonotes_charlm", verbose=False)
        assert sorted(os.listdir(test_dir)) == ['en', 'resources.json']
        en_dir = os.path.join(test_dir, 'en')
        en_dir_listing = sorted(os.listdir(en_dir))
        assert en_dir_listing == ['backward_charlm', 'forward_charlm', 'ner', 'pretrain']
        assert os.listdir(os.path.join(en_dir, 'ner')) == ['ontonotes_charlm.pt']
        for i in en_dir_listing:
            assert len(os.listdir(os.path.join(en_dir, i))) == 1


def test_download_two_models():
    """
    Test the download path for two NER models

    The package system should now allow for multiple NER models to be
    specified, and a consequence of that is it should be possible to
    download two models at once

    The expectation is that the two different NER models both download
    a different forward & backward charlm.  If that changes, the test
    will fail.  Best way to update it will be two different models
    which download two different charlms
    """
    with tempfile.TemporaryDirectory(dir=TEST_WORKING_DIR) as test_dir:
        stanza.download("en", model_dir=test_dir, processors="ner", package={"ner": ["ontonotes_charlm", "anatem"]}, verbose=False)
        assert sorted(os.listdir(test_dir)) == ['en', 'resources.json']
        en_dir = os.path.join(test_dir, 'en')
        en_dir_listing = sorted(os.listdir(en_dir))
        assert en_dir_listing == ['backward_charlm', 'forward_charlm', 'ner', 'pretrain']
        assert sorted(os.listdir(os.path.join(en_dir, 'ner'))) == ['anatem.pt', 'ontonotes_charlm.pt']
        for i in en_dir_listing:
            assert len(os.listdir(os.path.join(en_dir, i))) == 2


def test_process_pipeline_parameters():
    """
    Test a few options for specifying which processors to load
    """
    with tempfile.TemporaryDirectory(dir=TEST_WORKING_DIR) as test_dir:
        lang, model_dir, package, processors = common.process_pipeline_parameters("en", test_dir, None, "tokenize,pos")
        assert processors == {"tokenize": "default", "pos": "default"}
        assert package == None

        lang, model_dir, package, processors = common.process_pipeline_parameters("en", test_dir, {"tokenize": "spacy"}, "tokenize,pos")
        assert processors == {"tokenize": "spacy", "pos": "default"}
        assert package == None

        lang, model_dir, package, processors = common.process_pipeline_parameters("en", test_dir, {"pos": "ewt"}, "tokenize,pos")
        assert processors == {"tokenize": "default", "pos": "ewt"}
        assert package == None

        lang, model_dir, package, processors = common.process_pipeline_parameters("en", test_dir, "ewt", "tokenize,pos")
        assert processors == {"tokenize": "ewt", "pos": "ewt"}
        assert package == None

def test_language_resources():
    resources = common.load_resources_json(TEST_MODELS_DIR)

    # check that an unknown language comes back as None
    bad_lang = 'z'
    while bad_lang in resources and len(bad_lang) < 100:
        bad_lang = bad_lang + 'z'
    assert bad_lang not in resources
    assert common.get_language_resources(resources, bad_lang) == None

    # check the parameters of the test make sense
    # there should be 'zh' which is an alias of 'zh-hans'
    assert "zh" in resources
    assert "alias" in resources["zh"]
    assert resources["zh"]["alias"] == "zh-hans"

    # check that getting the resources for either 'zh' or 'zh-hans'
    # return the simplified Chinese resources
    zh_resources = common.get_language_resources(resources, "zh")
    assert "tokenize" in zh_resources
    assert "alias" not in zh_resources
    assert "Chinese" in zh_resources["lang_name"]

    zh_hans_resources = common.get_language_resources(resources, "zh-hans")
    assert zh_resources == zh_hans_resources



import itertools
import pytest

from stanza.models.constituency import parse_transitions
from stanza.models.constituency import tree_reader
from stanza.models.constituency.base_model import SimpleModel
from stanza.models.constituency.in_order_oracle import *
from stanza.models.constituency.parse_transitions import CloseConstituent, OpenConstituent, Shift, TransitionScheme
from stanza.models.constituency.transition_sequence import build_treebank

from stanza.tests import *
from stanza.tests.constituency.test_transition_sequence import reconstruct_tree

pytestmark = [pytest.mark.pipeline, pytest.mark.travis]

# A sample tree from PTB with a single unary transition (at a location other than root)
SINGLE_UNARY_TREE = """
( (S
    (NP-SBJ-1 (DT A) (NN record) (NN date) )
    (VP (VBZ has) (RB n't)
      (VP (VBN been)
        (VP (VBN set)
          (NP (-NONE- *-1) ))))
    (. .) ))
"""

#  [Shift, OpenConstituent(('NP-SBJ-1',)), Shift, Shift, CloseConstituent, OpenConstituent(('S',)), Shift, OpenConstituent(('VP',)), Shift, Shift, OpenConstituent(('VP',)), Shift, OpenConstituent(('VP',)), Shift, OpenConstituent(('NP',)), CloseConstituent, CloseConstituent, CloseConstituent, CloseConstituent, Shift, CloseConstituent, OpenConstituent(('ROOT',)), CloseConstituent]

# A sample tree from PTB with a double unary transition (at a location other than root)
DOUBLE_UNARY_TREE = """
( (S
    (NP-SBJ
      (NP (RB Not) (PDT all) (DT those) )
      (SBAR
        (WHNP-3 (WP who) )
        (S
          (NP-SBJ (-NONE- *T*-3) )
          (VP (VBD wrote) ))))
    (VP (VBP oppose)
      (NP (DT the) (NNS changes) ))
    (. .) ))
"""

# A sample tree from PTB with a triple unary transition (at a location other than root)
# The triple unary is at the START of the next bracket, which affects how the
# dynamic oracle repairs the transition sequence
TRIPLE_UNARY_START_TREE = """
( (S
    (PRN
      (S
        (NP-SBJ (-NONE- *) )
        (VP (VB See) )))
    (, ,)
    (NP-SBJ
      (NP (DT the) (JJ other) (NN rule) )
      (PP (IN of)
        (NP (NN thumb) ))
      (PP (IN about)
        (NP (NN ballooning) )))))
"""

# A sample tree from PTB with a triple unary transition (at a location other than root)
# The triple unary is at the END of the next bracket, which affects how the
# dynamic oracle repairs the transition sequence
TRIPLE_UNARY_END_TREE = """
( (S
    (NP (NNS optimists) )
    (VP (VBP expect) 
      (S 
        (NP-SBJ-4 (NNP Hong) (NNP Kong) )
        (VP (TO to) 
          (VP (VB hum) 
            (ADVP-CLR (RB along) )
            (SBAR-MNR (RB as) 
              (S 
                (NP-SBJ (-NONE- *-4) )
                (VP (-NONE- *?*) 
                  (ADVP-TMP (IN before) ))))))))))
"""

TREES = [SINGLE_UNARY_TREE, DOUBLE_UNARY_TREE, TRIPLE_UNARY_START_TREE, TRIPLE_UNARY_END_TREE]
TREEBANK = "\n".join(TREES)

NOUN_PHRASE_TREE = """
( (NP
    (NP (NNP Chicago) (POS 's))
    (NNP Goodman)
    (NNP Theatre)))
"""

WIDE_NP_TREE = """
( (S
    (NP-SBJ (DT These) (NNS studies))
    (VP (VBP demonstrate)
      (SBAR (IN that)
        (S
          (NP-SBJ (NNS mice))
          (VP (VBP are)
            (NP-PRD
              (NP (DT a)
                (ADJP (JJ practical)
                  (CC and)
                  (JJ powerful))
                (JJ experimental) (NN system))
              (SBAR
                (WHADVP-2 (-NONE- *0*))
                (S
                  (NP-SBJ (-NONE- *PRO*))
                  (VP (TO to)
                    (VP (VB study)
                      (NP (DT the) (NN genetics)))))))))))))
"""

WIDE_TREES = [NOUN_PHRASE_TREE, WIDE_NP_TREE]
WIDE_TREEBANK = "\n".join(WIDE_TREES)

ROOT_LABELS = ["ROOT"]

def get_repairs(gold_sequence, wrong_transition, repair_fn):
    """
    Use the repair function and the wrong transition to iterate over the gold sequence

    Returns a list of possible repairs, one for each position in the sequence
    Repairs are tuples, (idx, seq)
    """
    repairs = [(idx, repair_fn(gold_transition, wrong_transition, gold_sequence, idx, ROOT_LABELS, None, None))
               for idx, gold_transition in enumerate(gold_sequence)]
    repairs = [x for x in repairs if x[1] is not None]
    return repairs

@pytest.fixture(scope="module")
def unary_trees():
    trees = tree_reader.read_trees(TREEBANK)
    trees = [t.prune_none().simplify_labels() for t in trees]
    assert len(trees) == len(TREES)

    return trees

@pytest.fixture(scope="module")
def gold_sequences(unary_trees):
    gold_sequences = build_treebank(unary_trees, TransitionScheme.IN_ORDER)
    return gold_sequences

@pytest.fixture(scope="module")
def wide_trees():
    trees = tree_reader.read_trees(WIDE_TREEBANK)
    trees = [t.prune_none().simplify_labels() for t in trees]
    assert len(trees) == len(WIDE_TREES)

    return trees

def test_wrong_open_root(gold_sequences):
    """
    Test the results of the dynamic oracle on a few trees if the ROOT is mishandled.
    """
    wrong_transition = OpenConstituent("S")
    gold_transition = OpenConstituent("ROOT")
    close_transition = CloseConstituent()

    for gold_sequence in gold_sequences:
        # each of the sequences should be ended with ROOT, Close
        assert gold_sequence[-2] == gold_transition

        repairs = get_repairs(gold_sequence, wrong_transition, fix_wrong_open_root_error)
        # there is only spot in the sequence with a ROOT, so there should
        # be exactly one location which affords a S/ROOT replacement
        assert len(repairs) == 1
        repair = repairs[0]

        # the repair should occur at the -2 position, which is where ROOT is
        assert repair[0] == len(gold_sequence) - 2
        # and the resulting list should have the wrong transition followed by a Close
        # to give the model another chance to close the tree
        expected = gold_sequence[:-2] + [wrong_transition, close_transition] + gold_sequence[-2:]
        assert repair[1] == expected

def test_missed_unary(gold_sequences):
    """
    Test the repairs of an open/open error if it is effectively a skipped unary transition
    """
    wrong_transition = OpenConstituent("S")

    repairs = get_repairs(gold_sequences[0], wrong_transition, fix_wrong_open_unary_chain)
    assert len(repairs) == 0

    # here we are simulating picking NT-S instead of NT-VP
    # the DOUBLE_UNARY tree has one location where this is relevant, index 11
    repairs = get_repairs(gold_sequences[1], wrong_transition, fix_wrong_open_unary_chain)
    assert len(repairs) == 1
    assert repairs[0][0] == 11
    assert repairs[0][1] == gold_sequences[1][:11] + gold_sequences[1][13:]

    # the TRIPLE_UNARY_START tree has two locations where this is relevant
    # at index 1, the pattern goes (S (VP ...))
    # so choosing S instead of VP means you can skip the VP and only miss that one bracket
    # at index 5, the pattern goes (S (PRN (S (VP ...))) (...))
    # note that this is capturing a unary transition into a larger constituent
    # skipping the PRN is satisfactory
    repairs = get_repairs(gold_sequences[2], wrong_transition, fix_wrong_open_unary_chain)
    assert len(repairs) == 2
    assert repairs[0][0] == 1
    assert repairs[0][1] == gold_sequences[2][:1] + gold_sequences[2][3:]
    assert repairs[1][0] == 5
    assert repairs[1][1] == gold_sequences[2][:5] + gold_sequences[2][7:]

    # The TRIPLE_UNARY_END tree has 2 sections of tree for a total of 3 locations
    # where the repair might happen
    # Surprisingly the unary transition at the very start can only be
    # repaired by skipping it and using the outer S transition instead
    # The second repair overall (first repair in the second location)
    # should have a double skip to reach the S node
    repairs = get_repairs(gold_sequences[3], wrong_transition, fix_wrong_open_unary_chain)
    assert len(repairs) == 3
    assert repairs[0][0] == 1
    assert repairs[0][1] == gold_sequences[3][:1] + gold_sequences[3][3:]
    assert repairs[1][0] == 21
    assert repairs[1][1] == gold_sequences[3][:21] + gold_sequences[3][25:]
    assert repairs[2][0] == 23
    assert repairs[2][1] == gold_sequences[3][:23] + gold_sequences[3][25:]


def test_open_with_stuff(unary_trees, gold_sequences):
    wrong_transition = OpenConstituent("S")
    expected_trees = [
        "(ROOT (S (DT A) (NN record) (NN date) (VP (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set)))) (. .)))",
        "(ROOT (S (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))",
        None,
        "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NNP Hong) (NNP Kong) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before)))))))))))"
    ]

    for tree, gold_sequence, expected in zip(unary_trees, gold_sequences, expected_trees):
        repairs = get_repairs(gold_sequence, wrong_transition, fix_wrong_open_stuff_unary)
        if expected is None:
            assert len(repairs) == 0
        else:
            assert len(repairs) == 1
            result = reconstruct_tree(tree, repairs[0][1])
            assert str(result) == expected

def test_general_open(gold_sequences):
    wrong_transition = OpenConstituent("SBARQ")

    for sequence in gold_sequences:
        repairs = get_repairs(sequence, wrong_transition, fix_wrong_open_general)
        assert len(repairs) == sum(isinstance(x, OpenConstituent) for x in sequence) - 1
        for repair in repairs:
            assert len(repair[1]) == len(sequence)
            assert repair[1][repair[0]] == wrong_transition
            assert repair[1][:repair[0]] == sequence[:repair[0]]
            assert repair[1][repair[0]+1:] == sequence[repair[0]+1:]

def test_missed_unary(unary_trees, gold_sequences):
    shift_transition = Shift()
    close_transition = CloseConstituent()

    expected_close_results = [
        [(12, 2)],
        [(11, 4), (13, 2)],
        # (NP NN thumb) and (NP NN ballooning) are both candidates for this repair
        [(18, 2), (24, 2)],
        [(21, 6), (23, 4), (25, 2)],
    ]

    expected_shift_results = [
        (),
        (),
        (),
        # (ADVP-CLR (RB along)) is followed by a shift
        [(16, 2)],
    ]

    for tree, sequence, expected_close, expected_shift in zip(unary_trees, gold_sequences, expected_close_results, expected_shift_results):
        repairs = get_repairs(sequence, close_transition, fix_missed_unary)
        assert len(repairs) == len(expected_close)
        for repair, (expected_idx, expected_len) in zip(repairs, expected_close):
            assert repair[0] == expected_idx
            assert repair[1] == sequence[:expected_idx] + sequence[expected_idx+expected_len:]

        repairs = get_repairs(sequence, shift_transition, fix_missed_unary)
        assert len(repairs) == len(expected_shift)
        for repair, (expected_idx, expected_len) in zip(repairs, expected_shift):
            assert repair[0] == expected_idx
            assert repair[1] == sequence[:expected_idx] + sequence[expected_idx+expected_len:]

def test_open_shift(unary_trees, gold_sequences):
    shift_transition = Shift()

    expected_repairs = [
        [(7,  "(ROOT (S (NP (DT A) (NN record) (NN date)) (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set))) (. .)))"),
         (10, "(ROOT (S (NP (DT A) (NN record) (NN date)) (VP (VBZ has) (RB n't) (VBN been) (VP (VBN set))) (. .)))")],
        [(7,  "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (WP who) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"),
         (9,  "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (WHNP (WP who)) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"),
         (19, "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VBP oppose) (NP (DT the) (NNS changes)) (. .)))"),
         (21, "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VP (VBP oppose) (DT the) (NNS changes)) (. .)))")],
        [(14, "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning)))))"),
         (16, "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (IN of) (NP (NN thumb)) (PP (IN about) (NP (NN ballooning))))))"),
         (22, "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (IN about) (NP (NN ballooning)))))")],
        [(5,  "(ROOT (S (NP (NNS optimists)) (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (10, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (12, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (14, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (19, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (RB as) (S (VP (ADVP (IN before))))))))))")]
    ]

    for tree, sequence, expected in zip(unary_trees, gold_sequences, expected_repairs):
        repairs = get_repairs(sequence, shift_transition, fix_open_shift)
        assert len(repairs) == len(expected)
        for repair, (idx, expected_tree) in zip(repairs, expected):
            assert repair[0] == idx
            result_tree = reconstruct_tree(tree, repair[1])
            assert str(result_tree) == expected_tree


def test_open_close(unary_trees, gold_sequences):
    close_transition = CloseConstituent()

    expected_repairs = [
        [(7,  "(ROOT (S (S (NP (DT A) (NN record) (NN date)) (VBZ has)) (RB n't) (VP (VBN been) (VP (VBN set))) (. .)))"),
         (10, "(ROOT (S (NP (DT A) (NN record) (NN date)) (VP (VP (VBZ has) (RB n't) (VBN been)) (VP (VBN set))) (. .)))")],
        # missed the WHNP.  The surrounding SBAR cannot be created, either
        [(7, "(ROOT (S (NP (NP (NP (RB Not) (PDT all) (DT those)) (WP who)) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"),
         # missed the SBAR
         (9, "(ROOT (S (NP (NP (NP (RB Not) (PDT all) (DT those)) (WHNP (WP who))) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"),
         # missed the VP around "oppose the changes"
         (19, "(ROOT (S (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VBP oppose)) (NP (DT the) (NNS changes)) (. .)))"),
         # missed the NP in "the changes", looks pretty bad tbh
         (21, "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VP (VP (VBP oppose) (DT the)) (NNS changes)) (. .)))")],
        [(14, "(ROOT (S (S (PRN (S (VP (VB See)))) (, ,) (NP (DT the) (JJ other) (NN rule))) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning)))))"),
         (16, "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (NP (NP (DT the) (JJ other) (NN rule)) (IN of)) (NP (NN thumb)) (PP (IN about) (NP (NN ballooning))))))"),
         (22, "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (IN about)) (NP (NN ballooning)))))")],
        [(5, "(ROOT (S (S (NP (NNS optimists)) (VBP expect)) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (10, "(ROOT (S (NP (NNS optimists)) (VP (VP (VBP expect) (NP (NNP Hong) (NNP Kong))) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (12, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (S (NP (NNP Hong) (NNP Kong)) (TO to)) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (14, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (VP (TO to) (VB hum)) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))"),
         (19, "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VP (VB hum) (ADVP (RB along)) (RB as)) (S (VP (ADVP (IN before))))))))))")]
    ]

    for tree, sequence, expected in zip(unary_trees, gold_sequences, expected_repairs):
        repairs = get_repairs(sequence, close_transition, fix_open_close)

        assert len(repairs) == len(expected)
        for repair, (idx, expected_tree) in zip(repairs, expected):
            assert repair[0] == idx
            result_tree = reconstruct_tree(tree, repair[1])
            assert str(result_tree) == expected_tree

def test_shift_close(unary_trees, gold_sequences):
    """
    Test the fix for a shift -> close

    These errors can occur pretty much everywhere, and the fix is quite simple,
    so we only test a few cases.
    """

    close_transition = CloseConstituent()

    expected_tree = "(ROOT (S (NP (NP (DT A)) (NN record) (NN date)) (VP (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set)))) (. .)))"

    repairs = get_repairs(gold_sequences[0], close_transition, fix_shift_close)
    assert len(repairs) == 7
    result_tree = reconstruct_tree(unary_trees[0], repairs[0][1])
    assert str(result_tree) == expected_tree

    repairs = get_repairs(gold_sequences[1], close_transition, fix_shift_close)
    assert len(repairs) == 8

    repairs = get_repairs(gold_sequences[2], close_transition, fix_shift_close)
    assert len(repairs) == 8

    repairs = get_repairs(gold_sequences[3], close_transition, fix_shift_close)
    assert len(repairs) == 9
    for rep in repairs:
        if rep[0] == 16:
            # This one is special because it occurs as part of a unary
            # in other words, it should go unary, shift
            # and instead we are making it close where the unary should be
            # ... the unary would create "(ADVP (RB along))"
            expected_tree = "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VP (VB hum) (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before)))))))))))"
            result_tree = reconstruct_tree(unary_trees[3], rep[1])
            assert str(result_tree) == expected_tree
            break
    else:
        raise AssertionError("Did not find an expected repair location")

def test_close_open_shift_nested(unary_trees, gold_sequences):
    shift_transition = Shift()

    expected_trees = [{},
                      {4: "(ROOT (S (NP (RB Not) (PDT all) (DT those) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"},
                      {4: "(ROOT (S (VP (VB See)) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning))))))",
                       13: "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (DT the) (JJ other) (NN rule) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning))))))"},
                      {}]

    for tree, gold_sequence, expected in zip(unary_trees, gold_sequences, expected_trees):
        repairs = get_repairs(gold_sequence, shift_transition, fix_close_open_shift_nested)
        assert len(repairs) == len(expected)
        if len(expected) >= 1:
            for repair in repairs:
                assert repair[0] in expected.keys()
                result_tree = reconstruct_tree(tree, repair[1])
                assert str(result_tree) == expected[repair[0]]

def check_repairs(trees, gold_sequences, expected_trees, transition, repair_fn):
    for tree_idx, (gold_tree, gold_sequence, expected) in enumerate(zip(trees, gold_sequences, expected_trees)):
        repairs = get_repairs(gold_sequence, transition, repair_fn)
        if expected is not None:
            assert len(repairs) == len(expected)
            for repair in repairs:
                assert repair[0] in expected
                result_tree = reconstruct_tree(gold_tree, repair[1])
                assert str(result_tree) == expected[repair[0]]
        else:
            print("---------------------")
            print("{:P}".format(gold_tree))
            print(gold_sequence)
            #print(repairs)
            for repair in repairs:
                print("---------------------")
                print(gold_sequence)
                print(repair[1])
                result_tree = reconstruct_tree(gold_tree, repair[1])
                print("{:P}".format(gold_tree))
                print("{:P}".format(result_tree))
                print(tree_idx)
                print(repair[0])
                print(result_tree)

def test_close_open_shift_unambiguous(unary_trees, gold_sequences):
    shift_transition = Shift()

    expected_trees = [{},
                      {8: "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who) (S (VP (VBD wrote)))))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .)))"},
                      {},
                      {2: "(ROOT (S (NP (NNS optimists) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))))",
                       9: "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong) (VP (TO to) (VP (VB hum) (ADVP (RB along)) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))))"}]
    check_repairs(unary_trees, gold_sequences, expected_trees, shift_transition, fix_close_open_shift_unambiguous_bracket)

def test_close_open_shift_ambiguous_early(unary_trees, gold_sequences):
    shift_transition = Shift()

    expected_trees = [{4: "(ROOT (S (NP (DT A) (NN record) (NN date) (VP (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set))))) (. .)))"},
                      {16: "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes)))) (. .)))"},
                      {2: "(ROOT (S (PRN (S (VP (VB See) (, ,)))) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning))))))",
                       6: "(ROOT (S (PRN (S (VP (VB See))) (, ,)) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning))))))"},
                      {}]
    check_repairs(unary_trees, gold_sequences, expected_trees, shift_transition, fix_close_open_shift_ambiguous_bracket_early)

def test_close_open_shift_ambiguous_late(unary_trees, gold_sequences):
    shift_transition = Shift()

    expected_trees = [{4: "(ROOT (S (NP (DT A) (NN record) (NN date) (VP (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set)))) (. .))))"},
                      {16: "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote)))) (VP (VBP oppose) (NP (DT the) (NNS changes))) (. .))))"},
                      {2: "(ROOT (S (PRN (S (VP (VB See) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning)))))))))",
                       6: "(ROOT (S (PRN (S (VP (VB See))) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb))) (PP (IN about) (NP (NN ballooning)))))))"},
                      {}]
    check_repairs(unary_trees, gold_sequences, expected_trees, shift_transition, fix_close_open_shift_ambiguous_bracket_late)


def test_close_shift_shift(unary_trees, wide_trees):
    """
    Test that close -> shift works when there is a single block shifted after

    Includes a test specifically that there is no oracle action when there are two blocks after the missed close
    """
    shift_transition = Shift()

    expected_trees = [{15: "(ROOT (S (NP (DT A) (NN record) (NN date)) (VP (VBZ has) (RB n't) (VP (VBN been) (VP (VBN set))) (. .))))"},
                      {24: "(ROOT (S (NP (NP (RB Not) (PDT all) (DT those)) (SBAR (WHNP (WP who)) (S (VP (VBD wrote))))) (VP (VBP oppose) (NP (DT the) (NNS changes)) (. .))))"},
                      {20: "(ROOT (S (PRN (S (VP (VB See)))) (, ,) (NP (NP (DT the) (JJ other) (NN rule)) (PP (IN of) (NP (NN thumb)) (PP (IN about) (NP (NN ballooning)))))))"},
                      {17: "(ROOT (S (NP (NNS optimists)) (VP (VBP expect) (S (NP (NNP Hong) (NNP Kong)) (VP (TO to) (VP (VB hum) (ADVP (RB along) (SBAR (RB as) (S (VP (ADVP (IN before))))))))))))"},
                      {},
                      {}]

    test_trees = unary_trees + wide_trees
    gold_sequences = build_treebank(test_trees, TransitionScheme.IN_ORDER)

    check_repairs(test_trees, gold_sequences, expected_trees, shift_transition, fix_close_shift_shift_unambiguous)


def test_close_shift_shift_early(unary_trees, wide_trees):
    """
    Test that close -> shift works when there are multiple blocks shifted after

    Also checks that the single block case is skipped, so as to keep them separate when testing

    A tree with the expected property was specifically added for this test
    """
    shift_transition = Shift()

    test_trees = unary_trees + wide_trees
    gold_sequences = build_treebank(test_trees, TransitionScheme.IN_ORDER)

    expected_trees = [{},
                      {},
                      {},
                      {},
                      {},
                      {21: "(ROOT (S (NP (DT These) (NNS studies)) (VP (VBP demonstrate) (SBAR (IN that) (S (NP (NNS mice)) (VP (VBP are) (NP (NP (DT a) (ADJP (JJ practical) (CC and) (JJ powerful) (JJ experimental)) (NN system)) (SBAR (S (VP (TO to) (VP (VB study) (NP (DT the) (NN genetics)))))))))))))"}]

    check_repairs(test_trees, gold_sequences, expected_trees, shift_transition, fix_close_shift_shift_ambiguous_early)

def test_close_shift_shift_late(unary_trees, wide_trees):
    """
    Test that close -> shift works when there are multiple blocks shifted after

    Also checks that the single block case is skipped, so as to keep them separate when testing

    A tree with the expected property was specifically added for this test
    """
    shift_transition = Shift()

    test_trees = unary_trees + wide_trees
    gold_sequences = build_treebank(test_trees, TransitionScheme.IN_ORDER)

    expected_trees = [{},
                      {},
                      {},
                      {},
                      {},
                      {21: "(ROOT (S (NP (DT These) (NNS studies)) (VP (VBP demonstrate) (SBAR (IN that) (S (NP (NNS mice)) (VP (VBP are) (NP (NP (DT a) (ADJP (JJ practical) (CC and) (JJ powerful) (JJ experimental) (NN system))) (SBAR (S (VP (TO to) (VP (VB study) (NP (DT the) (NN genetics)))))))))))))"}]

    check_repairs(test_trees, gold_sequences, expected_trees, shift_transition, fix_close_shift_shift_ambiguous_late)



"""
Processes the SB10k dataset

The original description of the dataset and corpus_v1.0.tsv is here:

https://www.spinningbytes.com/resources/germansentiment/

Download script is here:

https://github.com/aritter/twitter_download

The problem with this file is that many of the tweets with labels no
longer exist.  Roughly 1/3 as of June 2020.

You can contact the authors for the complete dataset.

There is a paper describing some experiments run on the dataset here:
https://dl.acm.org/doi/pdf/10.1145/3038912.3052611
"""

import argparse
import os
import random

from enum import Enum

import stanza.utils.datasets.sentiment.process_utils as process_utils

class Split(Enum):
    TRAIN_DEV_TEST = 1
    TRAIN_DEV = 2
    TEST = 3

MAPPING = {'positive': "2",
           'neutral': "1",
           'negative': "0"}

def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('--csv_filename', type=str, default=None, help='CSV file to read in')
    parser.add_argument('--out_dir', type=str, default=None, help='Where to write the output files')
    parser.add_argument('--sentiment_column', type=int, default=2, help='Column with the sentiment')
    parser.add_argument('--text_column', type=int, default=3, help='Column with the text')
    parser.add_argument('--short_name', type=str, default="sb10k", help='short name to use when writing files')

    parser.add_argument('--split', type=lambda x: Split[x.upper()], default=Split.TRAIN_DEV_TEST,
                        help="How to split the resulting data")

    args = parser.parse_args(args=args)

    snippets = process_utils.read_snippets(args.csv_filename, args.sentiment_column, args.text_column, 'de', MAPPING)

    print(len(snippets))
    random.shuffle(snippets)

    os.makedirs(args.out_dir, exist_ok=True)
    if args.split is Split.TRAIN_DEV_TEST:
        process_utils.write_splits(args.out_dir,
                                   snippets,
                                   (process_utils.Split("%s.train.json" % args.short_name, 0.8),
                                    process_utils.Split("%s.dev.json" % args.short_name, 0.1),
                                    process_utils.Split("%s.test.json" % args.short_name, 0.1)))
    elif args.split is Split.TRAIN_DEV:
        process_utils.write_splits(args.out_dir,
                                   snippets,
                                   (process_utils.Split("%s.train.json" % args.short_name, 0.9),
                                    process_utils.Split("%s.dev.json" % args.short_name, 0.1)))
    elif args.split is Split.TEST:
        process_utils.write_list(os.path.join(args.out_dir, "%s.test.json" % args.short_name), snippets)
    else:
        raise ValueError("Unknown split method {}".format(args.split))

if __name__ == '__main__':
    random.seed(1234)
    main()




"""
Report the fraction of NER entities in one file which are present in another.

Purpose: show the coverage of one file on another, such as reporting
the number of entities in one dataset on another
"""


import argparse

from stanza.utils.datasets.ner.utils import read_json_entities

def parse_args():
    parser = argparse.ArgumentParser(description="Report the coverage of one NER file on another.")
    parser.add_argument('--train', type=str, nargs="+", required=True, help='File to use to collect the known entities (not necessarily train).')
    parser.add_argument('--test', type=str, nargs="+", required=True, help='File for which we want to know the ratio of known entities')
    args = parser.parse_args()
    return args

def report_known_entities(train_file, test_file):
    train_entities = read_json_entities(train_file)
    test_entities = read_json_entities(test_file)

    train_entities = set(x[0] for x in train_entities)
    total_score = sum(1 for x in test_entities if x[0] in train_entities)
    print(train_file, test_file, total_score / len(test_entities))

def main():
    args = parse_args()

    for train_idx, train_file in enumerate(args.train):
        if train_idx > 0:
            print()
        for test_file in args.test:
            report_known_entities(train_file, test_file)

if __name__ == '__main__':
    main()



"""
Tests that call a running CoreNLPClient.
"""

from http.server import BaseHTTPRequestHandler, HTTPServer
import multiprocessing
import pytest
import requests
import stanza.server as corenlp
import stanza.server.client as client
import shlex
import subprocess
import time

from stanza.models.constituency import tree_reader
from stanza.tests import *

# set the marker for this module
pytestmark = [pytest.mark.travis, pytest.mark.client]

TEXT = "Chris wrote a simple sentence that he parsed with Stanford CoreNLP.\n"

MAX_REQUEST_ATTEMPTS = 5

EN_GOLD = """
Sentence #1 (12 tokens):
Chris wrote a simple sentence that he parsed with Stanford CoreNLP.

Tokens:
[Text=Chris CharacterOffsetBegin=0 CharacterOffsetEnd=5 PartOfSpeech=NNP]
[Text=wrote CharacterOffsetBegin=6 CharacterOffsetEnd=11 PartOfSpeech=VBD]
[Text=a CharacterOffsetBegin=12 CharacterOffsetEnd=13 PartOfSpeech=DT]
[Text=simple CharacterOffsetBegin=14 CharacterOffsetEnd=20 PartOfSpeech=JJ]
[Text=sentence CharacterOffsetBegin=21 CharacterOffsetEnd=29 PartOfSpeech=NN]
[Text=that CharacterOffsetBegin=30 CharacterOffsetEnd=34 PartOfSpeech=WDT]
[Text=he CharacterOffsetBegin=35 CharacterOffsetEnd=37 PartOfSpeech=PRP]
[Text=parsed CharacterOffsetBegin=38 CharacterOffsetEnd=44 PartOfSpeech=VBD]
[Text=with CharacterOffsetBegin=45 CharacterOffsetEnd=49 PartOfSpeech=IN]
[Text=Stanford CharacterOffsetBegin=50 CharacterOffsetEnd=58 PartOfSpeech=NNP]
[Text=CoreNLP CharacterOffsetBegin=59 CharacterOffsetEnd=66 PartOfSpeech=NNP]
[Text=. CharacterOffsetBegin=66 CharacterOffsetEnd=67 PartOfSpeech=.]
""".strip()

def run_webserver(port, timeout_secs):
    class HTTPTimeoutHandler(BaseHTTPRequestHandler):
        def do_POST(self):
            time.sleep(timeout_secs)
            self.send_response(200)
            self.send_header('Content-type', 'text/plain; charset=utf-8')
            self.end_headers()
            self.wfile.write("HTTPMockServerTimeout")

    HTTPServer(('127.0.0.1', port), HTTPTimeoutHandler).serve_forever()

class HTTPMockServerTimeoutContext:
    """ For launching an HTTP server on certain port with an specified delay at responses """
    def __init__(self, port, timeout_secs):
        self.port = port
        self.timeout_secs = timeout_secs

    def __enter__(self):
        self.p = multiprocessing.Process(target=run_webserver, args=(self.port, self.timeout_secs))
        self.p.daemon = True
        self.p.start()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.p.terminate()

class TestCoreNLPClient:
    @pytest.fixture(scope="class")
    def corenlp_client(self):
        """ Client to run tests on """
        client = corenlp.CoreNLPClient(annotators='tokenize,ssplit,pos,lemma,ner,depparse',
                                       server_id='stanza_main_test_server')
        yield client
        client.stop()


    def test_connect(self, corenlp_client):
        corenlp_client.ensure_alive()
        assert corenlp_client.is_active
        assert corenlp_client.is_alive()


    def test_context_manager(self):
        with corenlp.CoreNLPClient(annotators="tokenize,ssplit",
                                   endpoint="http://localhost:9001") as context_client:
            ann = context_client.annotate(TEXT)
            assert corenlp.to_text(ann.sentence[0]) == TEXT[:-1]

    def test_no_duplicate_servers(self):
        """We expect a second server on the same port to fail"""
        with pytest.raises(corenlp.PermanentlyFailedException):
            with corenlp.CoreNLPClient(annotators="tokenize,ssplit") as duplicate_server:
                raise RuntimeError("This should have failed")

    def test_annotate(self, corenlp_client):
        ann = corenlp_client.annotate(TEXT)
        assert corenlp.to_text(ann.sentence[0]) == TEXT[:-1]


    def test_update(self, corenlp_client):
        ann = corenlp_client.annotate(TEXT)
        ann = corenlp_client.update(ann)
        assert corenlp.to_text(ann.sentence[0]) == TEXT[:-1]


    def test_tokensregex(self, corenlp_client):
        pattern = '([ner: PERSON]+) /wrote/ /an?/ []{0,3} /sentence|article/'
        matches = corenlp_client.tokensregex(TEXT, pattern)
        assert len(matches["sentences"]) == 1
        assert matches["sentences"][0]["length"] == 1
        assert matches == {
            "sentences": [{
                "0": {
                    "text": "Chris wrote a simple sentence",
                    "begin": 0,
                    "end": 5,
                    "1": {
                        "text": "Chris",
                        "begin": 0,
                        "end": 1
                    }},
                "length": 1
            },]}


    def test_semgrex(self, corenlp_client):
        pattern = '{word:wrote} >nsubj {}=subject >obj {}=object'
        matches = corenlp_client.semgrex(TEXT, pattern, to_words=True)
        assert matches == [
            {
                "text": "wrote",
                "begin": 1,
                "end": 2,
                "$subject": {
                    "text": "Chris",
                    "begin": 0,
                    "end": 1
                },
                "$object": {
                    "text": "sentence",
                    "begin": 4,
                    "end": 5
                },
                "sentence": 0,}]

    def test_tregex(self, corenlp_client):
        # the PP should be easy to parse
        pattern = 'PP < NP'
        matches = corenlp_client.tregex(TEXT, pattern)
        print(matches)
        assert matches == {
            'sentences': [
                {'0': {'sentIndex': 0, 'characterOffsetBegin': 45, 'codepointOffsetBegin': 45, 'characterOffsetEnd': 66, 'codepointOffsetEnd': 66,
                       'match': '(PP (IN with)\n  (NP (NNP Stanford) (NNP CoreNLP)))\n',
                       'spanString': 'with Stanford CoreNLP', 'namedNodes': []}}
            ]
        }

    def test_tregex_trees(self, corenlp_client):
        """
        Test the results of tregex run on trees w/o parsing
        """
        trees = tree_reader.read_trees("(ROOT (S (NP (NNP Jennifer)) (VP (VBZ has) (NP (JJ blue) (NN skin)))))   (ROOT (S (NP (PRP I)) (VP (VBP like) (NP (PRP$ her) (NNS antennae)))))")
        pattern = "VP < NP"
        matches = corenlp_client.tregex(pattern=pattern, trees=trees)
        assert matches == {
            'sentences': [
                {'0': {'sentIndex': 0, 'match': '(VP (VBZ has)\n  (NP (JJ blue) (NN skin)))\n', 'spanString': 'has blue skin', 'namedNodes': []}},
                {'0': {'sentIndex': 1, 'match': '(VP (VBP like)\n  (NP (PRP$ her) (NNS antennae)))\n', 'spanString': 'like her antennae', 'namedNodes': []}}
            ]
        }

    @pytest.fixture
    def external_server_9001(self):
        corenlp_home = client.resolve_classpath(None)
        start_cmd = f'java -Xmx5g -cp "{corenlp_home}" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -port 9001 ' \
                    f'-timeout 60000 -server_id stanza_external_server -serverProperties {SERVER_TEST_PROPS}'
        start_cmd = start_cmd and shlex.split(start_cmd)
        external_server_process = subprocess.Popen(start_cmd)

        yield external_server_process

        assert external_server_process
        external_server_process.terminate()
        external_server_process.wait(5)

    def test_external_server_legacy_start_server(self, external_server_9001):
        """ Test starting up an external server and accessing with a client with start_server=False """
        with corenlp.CoreNLPClient(start_server=False, endpoint="http://localhost:9001") as external_server_client:
            ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')
        assert ann.strip() == EN_GOLD

    def test_external_server_available(self, external_server_9001):
        """ Test starting up an external available server and accessing with a client with start_server=StartServer.DONT_START """
        time.sleep(5) # wait and make sure the external CoreNLP server is up and running
        with corenlp.CoreNLPClient(start_server=corenlp.StartServer.DONT_START, endpoint="http://localhost:9001") as external_server_client:
            ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')
        assert ann.strip() == EN_GOLD

    def test_external_server_unavailable(self):
        """ Test accessing with a client with start_server=StartServer.DONT_START to an external unavailable server """
        with pytest.raises(corenlp.AnnotationException):
            with corenlp.CoreNLPClient(start_server=corenlp.StartServer.DONT_START, endpoint="http://localhost:9001") as external_server_client:
                ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')

    def test_external_server_timeout(self):
        """ Test starting up an external server with long response time (20 seconds) and accessing with a client with start_server=StartServer.DONT_START and timeout=5000"""
        with HTTPMockServerTimeoutContext(9001, 20):
            time.sleep(5) # wait and make sure the external HTTPMockServer server is up and running
            with pytest.raises(corenlp.TimeoutException):
                with corenlp.CoreNLPClient(start_server=corenlp.StartServer.DONT_START, endpoint="http://localhost:9001", timeout=5000) as external_server_client:
                    ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')

    def test_external_server_try_start_with_external(self, external_server_9001):
        """ Test starting up an external server and accessing with a client with start_server=StartServer.TRY_START """
        time.sleep(5) # wait and make sure the external CoreNLP server is up and running
        with corenlp.CoreNLPClient(start_server=corenlp.StartServer.TRY_START,
                                   annotators='tokenize,ssplit,pos',
                                   endpoint="http://localhost:9001") as external_server_client:
            ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')
            assert external_server_client.server is None, "If this is not None, that indicates the client started a server instead of reusing an existing one"
        assert ann.strip() == EN_GOLD

    def test_external_server_try_start(self):
        """ Test starting up a server with a client with start_server=StartServer.TRY_START """
        with corenlp.CoreNLPClient(start_server=corenlp.StartServer.TRY_START,
                                   annotators='tokenize,ssplit,pos',
                                   endpoint="http://localhost:9001") as external_server_client:
            ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')
        assert ann.strip() == EN_GOLD

    def test_external_server_force_start(self, external_server_9001):
        """ Test starting up an external server and accessing with a client with start_server=StartServer.FORCE_START """
        time.sleep(5) # wait and make sure the external CoreNLP server is up and running
        with pytest.raises(corenlp.PermanentlyFailedException):
            with corenlp.CoreNLPClient(start_server=corenlp.StartServer.FORCE_START, endpoint="http://localhost:9001") as external_server_client:
                ann = external_server_client.annotate(TEXT, annotators='tokenize,ssplit,pos', output_format='text')



"""
Processes a kaggle covid-19 text classification dataset

The original description of the dataset is here:

https://www.kaggle.com/datasets/datatattle/covid-19-nlp-text-classification

There are two files in the archive, Corona_NLP_train.csv and Corona_NLP_test.csv
Unzip the files in archive.zip to $SENTIMENT_BASE/english/corona/Corona_NLP_train.csv

There is no dedicated dev set, so we randomly split train/dev
(using a specific seed, so that the split always comes out the same)
"""

import argparse
import os
import random

import stanza

import stanza.utils.datasets.sentiment.process_utils as process_utils
from stanza.utils.default_paths import get_default_paths

# TODO: could give an option to keep the 'extremely'
MAPPING = {'extremely positive': "2",
           'positive': "2",
           'neutral': "1",
           'negative': "0",
           'extremely negative': "0"}

def main(args=None):
    default_paths = get_default_paths()
    sentiment_base_dir = default_paths["SENTIMENT_BASE"]
    default_in_dir = os.path.join(sentiment_base_dir, "english", "corona")
    default_out_dir = default_paths["SENTIMENT_DATA_DIR"]

    parser = argparse.ArgumentParser()
    parser.add_argument('--in_dir', type=str, default=default_in_dir, help='Where to get the input files')
    parser.add_argument('--out_dir', type=str, default=default_out_dir, help='Where to write the output files')
    parser.add_argument('--short_name', type=str, default="en_corona", help='short name to use when writing files')
    args = parser.parse_args(args=args)

    TEXT_COLUMN = 4
    SENTIMENT_COLUMN = 5

    train_csv = os.path.join(args.in_dir, "Corona_NLP_train.csv")
    test_csv = os.path.join(args.in_dir, "Corona_NLP_test.csv")

    nlp = stanza.Pipeline("en", processors='tokenize')

    train_snippets = process_utils.read_snippets(train_csv, SENTIMENT_COLUMN, TEXT_COLUMN, 'en', MAPPING, delimiter=",", quotechar='"', skip_first_line=True, nlp=nlp, encoding="latin1")
    test_snippets = process_utils.read_snippets(test_csv, SENTIMENT_COLUMN, TEXT_COLUMN, 'en', MAPPING, delimiter=",", quotechar='"', skip_first_line=True, nlp=nlp, encoding="latin1")

    print("Read %d train snippets" % len(train_snippets))
    print("Read %d test snippets" % len(test_snippets))

    random.seed(1234)
    random.shuffle(train_snippets)

    os.makedirs(args.out_dir, exist_ok=True)
    process_utils.write_splits(args.out_dir,
                               train_snippets,
                               (process_utils.Split("%s.train.json" % args.short_name, 0.9),
                                process_utils.Split("%s.dev.json" % args.short_name, 0.1)))
    process_utils.write_list(os.path.join(args.out_dir, "%s.test.json" % args.short_name), test_snippets)

if __name__ == '__main__':
    main()




""" Describes AnaphicityScorer, a torch module that for a matrix of
mentions produces their anaphoricity scores.
"""
import torch

from stanza.models.coref import utils
from stanza.models.coref.config import Config


class AnaphoricityScorer(torch.nn.Module):
    """ Calculates anaphoricity scores by passing the inputs into a FFNN """
    def __init__(self,
                 in_features: int,
                 config: Config):
        super().__init__()
        hidden_size = config.hidden_size
        if not config.n_hidden_layers:
            hidden_size = in_features
        layers = []
        for i in range(config.n_hidden_layers):
            layers.extend([torch.nn.Linear(hidden_size if i else in_features,
                                           hidden_size),
                           torch.nn.LeakyReLU(),
                           torch.nn.Dropout(config.dropout_rate)])
        self.hidden = torch.nn.Sequential(*layers)
        self.out = torch.nn.Linear(hidden_size, out_features=1)

        # are we going to predict singletons
        self.predict_singletons = config.singletons

        if self.predict_singletons:
            # map to whether or not this is a start of a coref given all the
            # antecedents; not used when config.singletons = False because
            # we only need to know this for predicting singletons
            self.start_map = torch.nn.Linear(config.rough_k, out_features=1, bias=False)


    def forward(self, *,  # type: ignore  # pylint: disable=arguments-differ  #35566 in pytorch
                top_mentions: torch.Tensor,
                mentions_batch: torch.Tensor,
                pw_batch: torch.Tensor,
                top_rough_scores_batch: torch.Tensor,
                ) -> torch.Tensor:
        """ Builds a pairwise matrix, scores the pairs and returns the scores.

        Args:
            all_mentions (torch.Tensor): [n_mentions, mention_emb]
            mentions_batch (torch.Tensor): [batch_size, mention_emb]
            pw_batch (torch.Tensor): [batch_size, n_ants, pw_emb]
            top_indices_batch (torch.Tensor): [batch_size, n_ants]
            top_rough_scores_batch (torch.Tensor): [batch_size, n_ants]

        Returns:
            torch.Tensor [batch_size, n_ants + 1]
                anaphoricity scores for the pairs + a dummy column
        """
        # [batch_size, n_ants, pair_emb]
        pair_matrix = self._get_pair_matrix(mentions_batch, pw_batch, top_mentions)

        # [batch_size, n_ants] vs [batch_size, 1]
        # first is coref scores, the second is whether its the start of a coref
        if self.predict_singletons:
            scores, start = self._ffnn(pair_matrix)
            scores = utils.add_dummy(scores+top_rough_scores_batch, eps=True)

            return torch.cat([start, scores], dim=1)
        else:
            scores = self._ffnn(pair_matrix)
            return utils.add_dummy(scores+top_rough_scores_batch, eps=True)

    def _ffnn(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculates anaphoricity scores.

        Args:
            x: tensor of shape [batch_size, n_ants, n_features]

        Returns:
            tensor of shape [batch_size, n_ants]
        """
        x = self.out(self.hidden(x))
        x = x.squeeze(2)

        if not self.predict_singletons:
            return x

        # because sometimes we only have the first 49 anaphoricities
        start = x @ self.start_map.weight[:,:x.shape[1]].T
        return x, start

    @staticmethod
    def _get_pair_matrix(mentions_batch: torch.Tensor,
                         pw_batch: torch.Tensor,
                         top_mentions: torch.Tensor) -> torch.Tensor:
        """
        Builds the matrix used as input for AnaphoricityScorer.

        Args:
            all_mentions (torch.Tensor): [n_mentions, mention_emb],
                all the valid mentions of the document,
                can be on a different device
            mentions_batch (torch.Tensor): [batch_size, mention_emb],
                the mentions of the current batch,
                is expected to be on the current device
            pw_batch (torch.Tensor): [batch_size, n_ants, pw_emb],
                pairwise features of the current batch,
                is expected to be on the current device
            top_indices_batch (torch.Tensor): [batch_size, n_ants],
                indices of antecedents of each mention

        Returns:
            torch.Tensor: [batch_size, n_ants, pair_emb]
        """
        emb_size = mentions_batch.shape[1]
        n_ants = pw_batch.shape[1]

        a_mentions = mentions_batch.unsqueeze(1).expand(-1, n_ants, emb_size)
        b_mentions = top_mentions
        similarity = a_mentions * b_mentions

        out = torch.cat((a_mentions, b_mentions, similarity, pw_batch), dim=2)
        return out



import os

from stanza.models.lemma_classifier import evaluate_models
from stanza.models.lemma_classifier import train_lstm_model
from stanza.models.lemma_classifier import train_transformer_model
from stanza.models.lemma_classifier.constants import ModelType

from stanza.resources.default_packages import default_pretrains, TRANSFORMERS
from stanza.utils.training import common
from stanza.utils.training.common import Mode, add_charlm_args, build_lemma_charlm_args, choose_lemma_charlm, find_wordvec_pretrain

def add_lemma_args(parser):
    add_charlm_args(parser)

    parser.add_argument('--model_type', default=ModelType.LSTM, type=lambda x: ModelType[x.upper()],
                        help='Model type to use.  {}'.format(", ".join(x.name for x in ModelType)))

def build_model_filename(paths, short_name, command_args, extra_args):
    return os.path.join("saved_models", "lemma_classifier", short_name + "_lemma_classifier.pt")

def run_treebank(mode, paths, treebank, short_name, command_args, extra_args):
    short_language, dataset = short_name.split("_", 1)

    base_args = []
    if '--save_name' not in extra_args:
        base_args += ['--save_name', build_model_filename(paths, short_name, command_args, extra_args)]

    embedding_args = build_lemma_charlm_args(short_language, dataset, command_args.charlm)
    if '--wordvec_pretrain_file' not in extra_args:
        wordvec_pretrain = find_wordvec_pretrain(short_language, default_pretrains, {}, dataset)
        embedding_args += ["--wordvec_pretrain_file", wordvec_pretrain]

    bert_args = []
    if command_args.model_type is ModelType.TRANSFORMER:
        if '--bert_model' not in extra_args:
            if short_language in TRANSFORMERS:
                bert_args = ['--bert_model', TRANSFORMERS.get(short_language)]
            else:
                raise ValueError("--bert_model not specified, so cannot figure out which transformer to use for language %s" % short_language)

    extra_train_args = []
    if command_args.force:
        extra_train_args.append('--force')

    if mode == Mode.TRAIN:
        train_args = []
        if "--train_file" not in extra_args:
            train_file = os.path.join("data", "lemma_classifier", "%s.train.lemma" % short_name)
            train_args += ['--train_file', train_file]
        if "--eval_file" not in extra_args:
            eval_file = os.path.join("data", "lemma_classifier", "%s.dev.lemma" % short_name)
            train_args += ['--eval_file', eval_file]
        train_args = base_args + train_args + extra_args + extra_train_args

        if command_args.model_type == ModelType.LSTM:
            train_args = embedding_args + train_args
            train_lstm_model.main(train_args)
        else:
            model_type_args = ["--model_type", command_args.model_type.name.lower()]
            train_args = bert_args + model_type_args + train_args
            train_transformer_model.main(train_args)

    if mode == Mode.SCORE_DEV or mode == Mode.TRAIN:
        eval_args = []
        if "--eval_file" not in extra_args:
            eval_file = os.path.join("data", "lemma_classifier", "%s.dev.lemma" % short_name)
            eval_args += ['--eval_file', eval_file]
        model_type_args = ["--model_type", command_args.model_type.name.lower()]
        eval_args = bert_args + model_type_args + base_args + eval_args + embedding_args + extra_args
        evaluate_models.main(eval_args)

    if mode == Mode.SCORE_TEST or mode == Mode.TRAIN:
        eval_args = []
        if "--eval_file" not in extra_args:
            eval_file = os.path.join("data", "lemma_classifier", "%s.test.lemma" % short_name)
            eval_args += ['--eval_file', eval_file]
        model_type_args = ["--model_type", command_args.model_type.name.lower()]
        eval_args = bert_args + model_type_args + base_args + eval_args + embedding_args + extra_args
        evaluate_models.main(eval_args)

def main(args=None):
    common.main(run_treebank, "lemma_classifier", "lemma_classifier", add_lemma_args, sub_argparse=train_lstm_model.build_argparse(), build_model_filename=build_model_filename, choose_charlm_method=choose_lemma_charlm, args=args)


if __name__ == '__main__':
    main()



"""
Test the semgrex interface
"""

import pytest
import stanza
import stanza.server.semgrex as semgrex
from stanza.models.common.doc import Document
from stanza.protobuf import SemgrexRequest
from stanza.utils.conll import CoNLL

from stanza.tests import *

pytestmark = [pytest.mark.travis, pytest.mark.client]

TEST_ONE_SENTENCE = [[
    {
        "id": 1,
        "text": "Unban",
        "lemma": "unban",
        "upos": "VERB",
        "xpos": "VB",
        "feats": "Mood=Imp|VerbForm=Fin",
        "head": 0,
        "deprel": "root",
        "misc": "start_char=0|end_char=5"
    },
    {
        "id": 2,
        "text": "Mox",
        "lemma": "Mox",
        "upos": "PROPN",
        "xpos": "NNP",
        "feats": "Number=Sing",
        "head": 3,
        "deprel": "compound",
        "misc": "start_char=6|end_char=9"
    },
    {
        "id": 3,
        "text": "Opal",
        "lemma": "Opal",
        "upos": "PROPN",
        "xpos": "NNP",
        "feats": "Number=Sing",
        "head": 1,
        "deprel": "obj",
        "misc": "start_char=10|end_char=14",
        "ner": "GEM"
    },
    {
        "id": 4,
        "text": "!",
        "lemma": "!",
        "upos": "PUNCT",
        "xpos": ".",
        "head": 1,
        "deprel": "punct",
        "misc": "start_char=14|end_char=15"
    }]]

TEST_TWO_SENTENCES = [[
    {
      "id": 1,
      "text": "Unban",
      "lemma": "unban",
      "upos": "VERB",
      "xpos": "VB",
      "feats": "Mood=Imp|VerbForm=Fin",
      "head": 0,
      "deprel": "root",
      "misc": "start_char=0|end_char=5"
    },
    {
      "id": 2,
      "text": "Mox",
      "lemma": "Mox",
      "upos": "PROPN",
      "xpos": "NNP",
      "feats": "Number=Sing",
      "head": 3,
      "deprel": "compound",
      "misc": "start_char=6|end_char=9"
    },
    {
      "id": 3,
      "text": "Opal",
      "lemma": "Opal",
      "upos": "PROPN",
      "xpos": "NNP",
      "feats": "Number=Sing",
      "head": 1,
      "deprel": "obj",
      "misc": "start_char=10|end_char=14"
    },
    {
      "id": 4,
      "text": "!",
      "lemma": "!",
      "upos": "PUNCT",
      "xpos": ".",
      "head": 1,
      "deprel": "punct",
      "misc": "start_char=14|end_char=15"
    }],
    [{
      "id": 1,
      "text": "Unban",
      "lemma": "unban",
      "upos": "VERB",
      "xpos": "VB",
      "feats": "Mood=Imp|VerbForm=Fin",
      "head": 0,
      "deprel": "root",
      "misc": "start_char=16|end_char=21"
    },
    {
      "id": 2,
      "text": "Mox",
      "lemma": "Mox",
      "upos": "PROPN",
      "xpos": "NNP",
      "feats": "Number=Sing",
      "head": 3,
      "deprel": "compound",
      "misc": "start_char=22|end_char=25"
    },
    {
      "id": 3,
      "text": "Opal",
      "lemma": "Opal",
      "upos": "PROPN",
      "xpos": "NNP",
      "feats": "Number=Sing",
      "head": 1,
      "deprel": "obj",
      "misc": "start_char=26|end_char=30"
    },
    {
      "id": 4,
      "text": "!",
      "lemma": "!",
      "upos": "PUNCT",
      "xpos": ".",
      "head": 1,
      "deprel": "punct",
      "misc": "start_char=30|end_char=31"
    }]]

ONE_SENTENCE_DOC = Document(TEST_ONE_SENTENCE, "Unban Mox Opal!")
TWO_SENTENCE_DOC = Document(TEST_TWO_SENTENCES, "Unban Mox Opal! Unban Mox Opal!")


def check_response(response, response_len=1, semgrex_len=1, source_index=1, target_index=3, reln='obj'):
    assert len(response.result) == response_len
    assert len(response.result[0].result) == semgrex_len
    for semgrex_result in response.result[0].result:
        assert len(semgrex_result.match) == 1
        assert semgrex_result.match[0].matchIndex == source_index
        for match in semgrex_result.match:
            assert len(match.node) == 2
            assert match.node[0].name == 'source'
            assert match.node[0].matchIndex == source_index
            assert match.node[1].name == 'target'
            assert match.node[1].matchIndex == target_index
            assert len(match.reln) == 1
            assert match.reln[0].name == 'zzz'
            assert match.reln[0].reln == reln

def test_multi():
    with semgrex.Semgrex() as sem:
        response = sem.process(ONE_SENTENCE_DOC, "{}=source >obj=zzz {}=target")
        check_response(response)
        response = sem.process(ONE_SENTENCE_DOC, "{}=source >obj=zzz {}=target")
        check_response(response)
        response = sem.process(TWO_SENTENCE_DOC, "{}=source >obj=zzz {}=target")
        check_response(response, response_len=2)

def test_single_sentence():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{}=source >obj=zzz {}=target")
    check_response(response)

def test_two_semgrex():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{}=source >obj=zzz {}=target", "{}=source >obj=zzz {}=target")
    check_response(response, semgrex_len=2)

def test_two_sentences():
    response = semgrex.process_doc(TWO_SENTENCE_DOC, "{}=source >obj=zzz {}=target")
    check_response(response, response_len=2)

def test_word_attribute():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{word:Mox}=source <=zzz {word:Opal}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')

def test_lemma_attribute():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{lemma:Mox}=source <=zzz {lemma:Opal}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')

def test_xpos_attribute():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{tag:NNP}=source <=zzz {word:Opal}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{pos:NNP}=source <=zzz {word:Opal}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')

def test_upos_attribute():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{cpos:PROPN}=source <=zzz {word:Opal}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')

def test_ner_attribute():
    response = semgrex.process_doc(ONE_SENTENCE_DOC, "{cpos:PROPN}=source <=zzz {ner:GEM}=target")
    check_response(response, response_len=1, source_index=2, reln='compound')

def test_hand_built_request():
    """
    Essentially a test program: the result should be a response with
    one match, two named nodes, one named relation
    """
    request = SemgrexRequest()
    request.semgrex.append("{}=source >obj=zzz {}=target")
    query = request.query.add()

    for idx, word in enumerate(['Unban', 'Mox', 'Opal']):
        token = query.token.add()
        token.word = word
        token.value = word

        node = query.graph.node.add()
        node.sentenceIndex = 1
        node.index = idx+1

    edge = query.graph.edge.add()
    edge.source = 1
    edge.target = 3
    edge.dep = 'obj'

    edge = query.graph.edge.add()
    edge.source = 3
    edge.target = 2
    edge.dep = 'compound'

    response = semgrex.send_semgrex_request(request)
    check_response(response)

BLANK_DEPENDENCY_SENTENCE = """
# sent_id = weblog-juancole.com_juancole_20051126063000_ENG_20051126_063000-0007
# text = You wonder if he was manipulating the market with his bombing targets.
1	You	you	PRON	PRP	Case=Nom|Person=2|PronType=Prs	2	nsubj	_	_
2	wonder	wonder	VERB	VBP	Mood=Ind|Number=Sing|Person=2|Tense=Pres|VerbForm=Fin	1	_	_	_
3	if	if	SCONJ	IN	_	6	mark	_	_
4	he	he	PRON	PRP	Case=Nom|Gender=Masc|Number=Sing|Person=3|PronType=Prs	6	nsubj	_	_
5	was	be	AUX	VBD	Mood=Ind|Number=Sing|Person=3|Tense=Past|VerbForm=Fin	6	aux	_	_
6	manipulating	manipulate	VERB	VBG	Tense=Pres|VerbForm=Part	2	ccomp	_	_
7	the	the	DET	DT	Definite=Def|PronType=Art	8	det	_	_
8	market	market	NOUN	NN	Number=Sing	6	obj	_	_
9	with	with	ADP	IN	_	12	case	_	_
10	his	his	PRON	PRP$	Case=Gen|Gender=Masc|Number=Sing|Person=3|Poss=Yes|PronType=Prs	12	nmod:poss	_	_
11	bombing	bombing	NOUN	NN	Number=Sing	12	compound	_	_
12	targets	target	NOUN	NNS	Number=Plur	6	obl	_	SpaceAfter=No
13	.	.	PUNCT	.	_	2	punct	_	_
""".lstrip()


def test_blank_dependency():
    """
    A user / contributor sent a dependency file with blank dependency labels and twisted up roots
    """
    blank_dep_doc = CoNLL.conll2doc(input_str=BLANK_DEPENDENCY_SENTENCE)
    blank_dep_request = semgrex.build_request(blank_dep_doc, "{}=root <_=edge {}")
    response = semgrex.send_semgrex_request(blank_dep_request)
    assert len(response.result) == 1
    assert len(response.result[0].result) == 1
    assert len(response.result[0].result[0].match) == 1
    # there should be a named node...
    assert len(response.result[0].result[0].match[0].node) == 1
    assert response.result[0].result[0].match[0].node[0].name == 'root'
    assert response.result[0].result[0].match[0].node[0].matchIndex == 2

    # ... and a named edge
    assert len(response.result[0].result[0].match[0].edge) == 1
    assert response.result[0].result[0].match[0].edge[0].source == 1
    assert response.result[0].result[0].match[0].edge[0].target == 2
    assert response.result[0].result[0].match[0].edge[0].reln == "_"



"""
A few tests of specific operations from the Dataset
"""

import os
import pytest

from stanza.models.common.doc import *
from stanza.models import tagger
from stanza.models.pos.data import Dataset, ShuffledDataset
from stanza.utils.conll import CoNLL

from stanza.tests.pos.test_tagger import TRAIN_DATA, TRAIN_DATA_NO_XPOS, TRAIN_DATA_NO_UPOS, TRAIN_DATA_NO_FEATS

def test_basic_reading():
    """
    Test that a dataset with no xpos is detected by the Dataset
    """
    # empty args for building the data object
    args = tagger.parse_args(args=[])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA)

    data = Dataset(train_doc, args, None)
    assert data.has_upos
    assert data.has_xpos
    assert data.has_feats

def test_no_xpos():
    """
    Test that a dataset with no xpos is detected by the Dataset
    """
    # empty args for building the data object
    args = tagger.parse_args(args=[])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA_NO_XPOS)

    data = Dataset(train_doc, args, None)
    assert data.has_upos
    assert not data.has_xpos
    assert data.has_feats

def test_no_upos():
    """
    Test that a dataset with no upos is detected by the Dataset
    """
    # empty args for building the data object
    args = tagger.parse_args(args=[])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA_NO_UPOS)

    data = Dataset(train_doc, args, None)
    assert not data.has_upos
    assert data.has_xpos
    assert data.has_feats

def test_no_feats():
    """
    Test that a dataset with no feats is detected by the Dataset
    """
    # empty args for building the data object
    args = tagger.parse_args(args=[])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA_NO_FEATS)

    data = Dataset(train_doc, args, None)
    assert data.has_upos
    assert data.has_xpos
    assert not data.has_feats

def test_no_augment():
    """
    Test that with no punct removing augmentation, the doc always has punct at the end
    """
    args = tagger.parse_args(args=["--shorthand", "en_test", "--augment_nopunct", "0.0"])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA)
    data = Dataset(train_doc, args, None)
    data = data.to_loader(batch_size=2)

    for i in range(50):
        for batch in data:
            for text in batch.text:
                assert text[-1] in (".", "!")

def test_augment():
    """
    Test that with 100% punct removing augmentation, the doc never has punct at the end
    """
    args = tagger.parse_args(args=["--shorthand", "en_test", "--augment_nopunct", "1.0"])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA)
    data = Dataset(train_doc, args, None)
    data = data.to_loader(batch_size=2)

    for i in range(50):
        for batch in data:
            for text in batch.text:
                assert text[-1] not in (".", "!")

def test_sometimes_augment():
    """
    Test 50% punct removing augmentation

    With this frequency, we should get a reasonable number of docs
    with a punct at the end and a reasonable without.
    """
    args = tagger.parse_args(args=["--shorthand", "en_test", "--augment_nopunct", "0.5"])

    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA)
    data = Dataset(train_doc, args, None)
    data = data.to_loader(batch_size=2)

    count_with = 0
    count_without = 0
    for i in range(50):
        for batch in data:
            for text in batch.text:
                if text[-1] in (".", "!"):
                    count_with += 1
                else:
                    count_without += 1

    # this should never happen
    # literally less than 1 in 10^20th odds
    assert count_with > 5
    assert count_without > 5


NO_XPOS_TEMPLATE = """
# text = Noxpos {indexp}
# sent_id = {index}
1	Noxpos	noxpos	NOUN	_	Number=Sing	0	root	_	start_char=0|end_char=8|ner=O
2	{indexp}	{indexp}	NUM	_	NumForm=Digit|NumType=Card	1	dep	_	start_char=9|end_char=10|ner=S-CARDINAL
""".strip()

YES_XPOS_TEMPLATE = """
# text = Yesxpos {indexp}
# sent_id = {index}
1	Yesxpos	yesxpos	NOUN	NN	Number=Sing	0	root	_	start_char=0|end_char=8|ner=O
2	{indexp}	{indexp}	NUM	CD	NumForm=Digit|NumType=Card	1	dep	_	start_char=9|end_char=10|ner=S-CARDINAL
""".strip()

def test_shuffle(tmp_path):
    args = tagger.parse_args(args=["--batch_size", "10", "--shorthand", "en_test", "--augment_nopunct", "0.0"])

    # 100 looked nice but was actually a 1/1000000 chance of the test failing
    # so let's crank it up to 1000 and make it 1/10^58
    no_xpos = [NO_XPOS_TEMPLATE.format(index=idx, indexp=idx+1) for idx in range(1000)]
    no_doc = CoNLL.conll2doc(input_str="\n\n".join(no_xpos))
    no_data = Dataset(no_doc, args, None)

    yes_xpos = [YES_XPOS_TEMPLATE.format(index=idx, indexp=idx+101) for idx in range(1000)]
    yes_doc = CoNLL.conll2doc(input_str="\n\n".join(yes_xpos))
    yes_data = Dataset(yes_doc, args, None)

    shuffled = ShuffledDataset([no_data, yes_data], 10)

    assert sum(1 for _ in shuffled) == 200

    num_with = 0
    num_without = 0
    for batch in shuffled:
        if batch.xpos is not None:
            num_with += 1
        else:
            num_without += 1
        # at the halfway point of the iteration, there should be at
        # least one in each category
        # for example, if we had forgotten to shuffle, this assertion would fail
        if num_with + num_without == 100:
            assert num_with > 1
            assert num_without > 1

    assert num_with == 100
    assert num_without == 100


EWT_SAMPLE = """
# sent_id = weblog-blogspot.com_alaindewitt_20040929103700_ENG_20040929_103700-0048
# text = Bush asked for permission to go to Alabama to work on a Senate campaign.
1	Bush	Bush	PROPN	NNP	Number=Sing	2	nsubj	2:nsubj	_
2	asked	ask	VERB	VBD	Mood=Ind|Number=Sing|Person=3|Tense=Past|VerbForm=Fin	0	root	0:root	_
3	for	for	ADP	IN	_	4	case	4:case	_
4	permission	permission	NOUN	NN	Number=Sing	2	obl	2:obl:for	_
5	to	to	PART	TO	_	6	mark	6:mark	_
6	go	go	VERB	VB	VerbForm=Inf	4	acl	4:acl:to	_
7	to	to	ADP	IN	_	8	case	8:case	_
8	Alabama	Alabama	PROPN	NNP	Number=Sing	6	obl	6:obl:to	_
9	to	to	PART	TO	_	10	mark	10:mark	_
10	work	work	VERB	VB	VerbForm=Inf	6	advcl	6:advcl:to	_
11	on	on	ADP	IN	_	14	case	14:case	_
12	a	a	DET	DT	Definite=Ind|PronType=Art	14	det	14:det	_
13	Senate	Senate	PROPN	NNP	Number=Sing	14	compound	14:compound	_
14	campaign	campaign	NOUN	NN	Number=Sing	10	obl	10:obl:on	SpaceAfter=No
15	.	.	PUNCT	.	_	2	punct	2:punct	_

# sent_id = weblog-blogspot.com_alaindewitt_20040929103700_ENG_20040929_103700-0049
# text = His superior officers said OK.
1	His	his	PRON	PRP$	Case=Gen|Gender=Masc|Number=Sing|Person=3|Poss=Yes|PronType=Prs	3	nmod:poss	3:nmod:poss	_
2	superior	superior	ADJ	JJ	Degree=Pos	3	amod	3:amod	_
3	officers	officer	NOUN	NNS	Number=Plur	4	nsubj	4:nsubj	_
4	said	say	VERB	VBD	Mood=Ind|Number=Plur|Person=3|Tense=Past|VerbForm=Fin	0	root	0:root	_
5	OK	ok	INTJ	UH	_	4	obj	4:obj	SpaceAfter=No
6	.	.	PUNCT	.	_	4	punct	4:punct	_

# sent_id = weblog-blogspot.com_alaindewitt_20040929103700_ENG_20040929_103700-0053
# text = In ’72 or ’73, if you were a pilot, active or Guard, and you had an obligation and wanted to get out, no problem.
1	In	in	ADP	IN	_	2	case	2:case	_
2	’72	'72	NUM	CD	NumForm=Digit|NumType=Card	10	obl	10:obl:in	_
3	or	or	CCONJ	CC	_	4	cc	4:cc	_
4	’73	'73	NUM	CD	NumForm=Digit|NumType=Card	2	conj	2:conj:or|10:obl:in	SpaceAfter=No
5	,	,	PUNCT	,	_	2	punct	2:punct	_
6	if	if	SCONJ	IN	_	10	mark	10:mark	_
7	you	you	PRON	PRP	Case=Nom|Person=2|PronType=Prs	10	nsubj	10:nsubj	_
8	were	be	AUX	VBD	Mood=Ind|Number=Sing|Person=2|Tense=Past|VerbForm=Fin	10	cop	10:cop	_
9	a	a	DET	DT	Definite=Ind|PronType=Art	10	det	10:det	_
10	pilot	pilot	NOUN	NN	Number=Sing	28	advcl	28:advcl:if	SpaceAfter=No
11	,	,	PUNCT	,	_	12	punct	12:punct	_
12	active	active	ADJ	JJ	Degree=Pos	10	amod	10:amod	_
13	or	or	CCONJ	CC	_	14	cc	14:cc	_
14	Guard	Guard	PROPN	NNP	Number=Sing	12	conj	10:amod|12:conj:or	SpaceAfter=No
15	,	,	PUNCT	,	_	18	punct	18:punct	_
16	and	and	CCONJ	CC	_	18	cc	18:cc	_
17	you	you	PRON	PRP	Case=Nom|Person=2|PronType=Prs	18	nsubj	18:nsubj|22:nsubj|24:nsubj:xsubj	_
18	had	have	VERB	VBD	Mood=Ind|Number=Sing|Person=2|Tense=Past|VerbForm=Fin	10	conj	10:conj:and|28:advcl:if	_
19	an	a	DET	DT	Definite=Ind|PronType=Art	20	det	20:det	_
20	obligation	obligation	NOUN	NN	Number=Sing	18	obj	18:obj	_
21	and	and	CCONJ	CC	_	22	cc	22:cc	_
22	wanted	want	VERB	VBD	Mood=Ind|Number=Sing|Person=2|Tense=Past|VerbForm=Fin	18	conj	18:conj:and	_
23	to	to	PART	TO	_	24	mark	24:mark	_
24	get	get	VERB	VB	VerbForm=Inf	22	xcomp	22:xcomp	_
25	out	out	ADV	RB	_	24	advmod	24:advmod	SpaceAfter=No
26	,	,	PUNCT	,	_	10	punct	10:punct	_
27	no	no	DET	DT	PronType=Neg	28	det	28:det	_
28	problem	problem	NOUN	NN	Number=Sing	0	root	0:root	SpaceAfter=No
29	.	.	PUNCT	.	_	28	punct	28:punct	_

# sent_id = weblog-blogspot.com_alaindewitt_20040929103700_ENG_20040929_103700-0054
# text = In fact, you were helping them solve their problem.”
1	In	in	ADP	IN	_	2	case	2:case	_
2	fact	fact	NOUN	NN	Number=Sing	6	obl	6:obl:in	SpaceAfter=No
3	,	,	PUNCT	,	_	2	punct	2:punct	_
4	you	you	PRON	PRP	Case=Nom|Person=2|PronType=Prs	6	nsubj	6:nsubj	_
5	were	be	AUX	VBD	Mood=Ind|Number=Sing|Person=2|Tense=Past|VerbForm=Fin	6	aux	6:aux	_
6	helping	help	VERB	VBG	Tense=Pres|VerbForm=Part	0	root	0:root	_
7	them	they	PRON	PRP	Case=Acc|Number=Plur|Person=3|PronType=Prs	6	obj	6:obj|8:nsubj:xsubj	_
8	solve	solve	VERB	VB	VerbForm=Inf	6	xcomp	6:xcomp	_
9	their	their	PRON	PRP$	Case=Gen|Number=Plur|Person=3|Poss=Yes|PronType=Prs	10	nmod:poss	10:nmod:poss	_
10	problem	problem	NOUN	NN	Number=Sing	8	obj	8:obj	SpaceAfter=No
11	.	.	PUNCT	.	_	6	punct	6:punct	SpaceAfter=No
12	”	"	PUNCT	''	_	6	punct	6:punct	_

# sent_id = weblog-blogspot.com_alaindewitt_20040929103700_ENG_20040929_103700-0055
# text = So Bush stopped flying.
1	So	so	ADV	RB	_	3	advmod	3:advmod	_
2	Bush	Bush	PROPN	NNP	Number=Sing	3	nsubj	3:nsubj|4:nsubj:xsubj	_
3	stopped	stop	VERB	VBD	Mood=Ind|Number=Sing|Person=3|Tense=Past|VerbForm=Fin	0	root	0:root	_
4	flying	fly	VERB	VBG	VerbForm=Ger	3	xcomp	3:xcomp	SpaceAfter=No
5	.	.	PUNCT	.	_	3	punct	3:punct	_
""".lstrip()

def test_length_limited_dataloader():
    sample = CoNLL.conll2doc(input_str=EWT_SAMPLE)

    args = tagger.parse_args(args=["--batch_size", "10", "--shorthand", "en_test", "--augment_nopunct", "0.0"])
    data = Dataset(sample, args, None)

    # this should read the whole dataset
    dl = data.to_length_limited_loader(5, 1000)
    batches = [batch.idx for batch in dl]
    assert batches == [(0, 1, 2, 3, 4)]

    dl = data.to_length_limited_loader(4, 1000)
    batches = [batch.idx for batch in dl]
    assert batches == [(0, 1, 2, 3), (4,)]

    dl = data.to_length_limited_loader(2, 1000)
    batches = [batch.idx for batch in dl]
    assert batches == [(0, 1), (2, 3), (4,)]

    # the first three sentences should reach this limit
    dl = data.to_length_limited_loader(5, 55)
    batches = [batch.idx for batch in dl]
    assert batches == [(0, 1, 2), (3, 4)]

    # the third sentence (2) is already past this limit by itself
    dl = data.to_length_limited_loader(5, 25)
    batches = [batch.idx for batch in dl]
    assert batches == [(0, 1), (2,), (3, 4)]



"""
Test a couple basic data functions, such as processing a doc for its lemmas
"""

import pytest

from stanza.models.common.doc import Document
from stanza.models.lemma.data import DataLoader
from stanza.utils.conll import CoNLL

TRAIN_DATA = """
# sent_id = weblog-juancole.com_juancole_20051126063000_ENG_20051126_063000-0003
# text = DPA: Iraqi authorities announced that they had busted up 3 terrorist cells operating in Baghdad.
1	DPA	DPA	PROPN	NNP	Number=Sing	0	root	0:root	SpaceAfter=No
2	:	:	PUNCT	:	_	1	punct	1:punct	_
3	Iraqi	Iraqi	ADJ	JJ	Degree=Pos	4	amod	4:amod	_
4	authorities	authority	NOUN	NNS	Number=Plur	5	nsubj	5:nsubj	_
5	announced	announce	VERB	VBD	Mood=Ind|Number=Plur|Person=3|Tense=Past|VerbForm=Fin	1	parataxis	1:parataxis	_
6	that	that	SCONJ	IN	_	9	mark	9:mark	_
7	they	they	PRON	PRP	Case=Nom|Number=Plur|Person=3|PronType=Prs	9	nsubj	9:nsubj	_
8	had	have	AUX	VBD	Mood=Ind|Number=Plur|Person=3|Tense=Past|VerbForm=Fin	9	aux	9:aux	_
9	busted	bust	VERB	VBN	Tense=Past|VerbForm=Part	5	ccomp	5:ccomp	_
10	up	up	ADP	RP	_	9	compound:prt	9:compound:prt	_
11	3	3	NUM	CD	NumForm=Digit|NumType=Card	13	nummod	13:nummod	_
12	terrorist	terrorist	ADJ	JJ	Degree=Pos	13	amod	13:amod	_
13	cells	cell	NOUN	NNS	Number=Plur	9	obj	9:obj	_
14	operating	operate	VERB	VBG	VerbForm=Ger	13	acl	13:acl	_
15	in	in	ADP	IN	_	16	case	16:case	_
16	Baghdad	Baghdad	PROPN	NNP	Number=Sing	14	obl	14:obl:in	SpaceAfter=No
17	.	.	PUNCT	.	_	1	punct	1:punct	_

# sent_id = weblog-juancole.com_juancole_20051126063000_ENG_20051126_063000-0004
# text = Two of them were being run by 2 officials of the Ministry of the Interior!
1	Two	two	NUM	CD	NumForm=Word|NumType=Card	6	nsubj:pass	6:nsubj:pass	_
2	of	of	ADP	IN	_	3	case	3:case	_
3	them	they	PRON	PRP	Case=Acc|Number=Plur|Person=3|PronType=Prs	1	nmod	1:nmod:of	_
4	were	be	AUX	VBD	Mood=Ind|Number=Plur|Person=3|Tense=Past|VerbForm=Fin	6	aux	6:aux	_
5	being	be	AUX	VBG	VerbForm=Ger	6	aux:pass	6:aux:pass	_
6	run	run	VERB	VBN	Tense=Past|VerbForm=Part|Voice=Pass	0	root	0:root	_
7	by	by	ADP	IN	_	9	case	9:case	_
8	2	2	NUM	CD	NumForm=Digit|NumType=Card	9	nummod	9:nummod	_
9	officials	official	NOUN	NNS	Number=Plur	6	obl	6:obl:by	_
10	of	of	ADP	IN	_	12	case	12:case	_
11	the	the	DET	DT	Definite=Def|PronType=Art	12	det	12:det	_
12	Ministry	Ministry	PROPN	NNP	Number=Sing	9	nmod	9:nmod:of	_
13	of	of	ADP	IN	_	15	case	15:case	_
14	the	the	DET	DT	Definite=Def|PronType=Art	15	det	15:det	_
15	Interior	Interior	PROPN	NNP	Number=Sing	12	nmod	12:nmod:of	SpaceAfter=No
16	!	!	PUNCT	.	_	6	punct	6:punct	_

""".lstrip()

GOESWITH_DATA = """
# sent_id = email-enronsent27_01-0041
# newpar id = email-enronsent27_01-p0005
# text = Ken Rice@ENRON COMMUNICATIONS
1	Ken	kenrice@enroncommunications	X	GW	Typo=Yes	0	root	0:root	_
2	Rice@ENRON	_	X	GW	_	1	goeswith	1:goeswith	_
3	COMMUNICATIONS	_	X	ADD	_	1	goeswith	1:goeswith	_

""".lstrip()

CORRECT_FORM_DATA = """
# sent_id = weblog-blogspot.com_healingiraq_20040409053012_ENG_20040409_053012-0019
# text = They are targetting ambulances
1	They	they	PRON	PRP	Case=Nom|Number=Plur|Person=3|PronType=Prs	3	nsubj	3:nsubj	_
2	are	be	AUX	VBP	Mood=Ind|Number=Plur|Person=3|Tense=Pres|VerbForm=Fin	3	aux	3:aux	_
3	targetting	target	VERB	VBG	Tense=Pres|Typo=Yes|VerbForm=Part	0	root	0:root	CorrectForm=targeting
4	ambulances	ambulance	NOUN	NNS	Number=Plur	3	obj	3:obj	SpaceAfter=No
"""

BLANKS_DATA = """
# sent_id = weblog-juancole.com_juancole_20051126063000_ENG_20051126_063000-0018
# text = Guerrillas killed an engineer, Asi Ali, from Tikrit.
1	Guerrillas	_	NOUN	NNS	Number=Plur	2	nsubj	2:nsubj	_
2	killed	_	VERB	VBD	Mood=Ind|Number=Plur|Person=3|Tense=Past|VerbForm=Fin	0	root	0:root	_
3	an	a	DET	DT	Definite=Ind|PronType=Art	4	det	4:det	_
4	engineer	_	NOUN	NN	Number=Sing	2	obj	2:obj	SpaceAfter=No

""".lstrip()


def test_load_document():
    train_doc = CoNLL.conll2doc(input_str=TRAIN_DATA)
    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=True)
    assert len(data) == 33 # meticulously counted by hand
    assert all(len(x) == 3 for x in data)

    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=False)
    assert len(data) == 33
    assert all(len(x) == 3 for x in data)

def test_load_goeswith():
    raw_data = TRAIN_DATA + GOESWITH_DATA
    train_doc = CoNLL.conll2doc(input_str=raw_data)
    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=True)
    assert len(data) == 36 # will be the same as in test_load_document with three additional words
    assert all(len(x) == 3 for x in data)

    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=False)
    assert len(data) == 33 # will be the same as in test_load_document, but with the trailing 3 GOESWITH removed
    assert all(len(x) == 3 for x in data)

def test_correct_form():
    raw_data = TRAIN_DATA + CORRECT_FORM_DATA
    train_doc = CoNLL.conll2doc(input_str=raw_data)
    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=True)
    assert len(data) == 37
    # the 'targeting' correction should not be applied if evaluation=True
    # when evaluation=False, then the CorrectForms will be applied
    assert not any(x[0] == 'targeting' for x in data)

    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=False)
    assert len(data) == 38 # the same, but with an extra row so the model learns both 'targetting' and 'targeting'
    assert any(x[0] == 'targeting' for x in data)
    assert any(x[0] == 'targetting' for x in data)

def test_load_blank():
    raw_data = TRAIN_DATA + BLANKS_DATA
    train_doc = CoNLL.conll2doc(input_str=raw_data)
    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=False, evaluation=False)
    assert len(data) == 37 # will be the same as in test_load_document with FOUR additional words
    assert all(len(x) == 3 for x in data)

    data = DataLoader.load_doc(train_doc, caseless=False, skip_blank_lemmas=True, evaluation=False)
    assert len(data) == 34 # will be the same as in test_load_document, but one extra word is added.  others were blank
    assert all(len(x) == 3 for x in data)




"""
Given two ensembles and a tokenized file, output the trees for which those ensembles agree and report how many of the sub-models agree on those trees.

For example:

python3 -m stanza.utils.datasets.constituency.build_silver_dataset --tokenized_file /u/nlp/data/constituency-parser/italian/2024_wiki_tokenization/it_wiki_tokenized_AA.txt --lang it --output_file asdf.out --e1 saved_models/constituency/it_vit_electra_100?_top_constituency.pt --e2 saved_models/constituency/it_vit_electra_100?_constituency.pt

for i in `echo f g h i j k l m n o p q r s t`; do nlprun -d a6000 "python3 -m stanza.utils.datasets.constituency.build_silver_dataset --tokenized_file /u/nlp/data/constituency-parser/italian/2024_wiki_tokenization/it_wiki_tok_6M_a$i.txt --lang it --output_file /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/a$i.trees --e1 saved_models/constituency/it_vit_electra_100?_top_constituency.pt --e2 saved_models/constituency/it_vit_electra_100?_constituency.pt" -o /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/a$i.out; done

for i in `echo a b c d`; do nlprun -d a6000 "python3 -m stanza.utils.datasets.constituency.build_silver_dataset --tokenized_file /u/nlp/data/constituency-parser/english/en_wiki_2023/shuf_1M.a$i --lang en --output_file /u/nlp/data/constituency-parser/english/2024_en_ptb3_electra/forward_a$i.trees --e1 saved_models/constituency/en_ptb3_electra-large_100?_in_constituency.pt --e2 saved_models/constituency/en_ptb3_electra-large_100?_top_constituency.pt" -o /u/nlp/data/constituency-parser/english/2024_en_ptb3_electra/forward_a$i.out; done
"""

import argparse
import json

import logging

from stanza.models.common import utils
from stanza.models.common.foundation_cache import FoundationCache
from stanza.models.constituency import retagging
from stanza.models.constituency import text_processing
from stanza.models.constituency import tree_reader
from stanza.models.constituency.ensemble import Ensemble
from stanza.utils.get_tqdm import get_tqdm

tqdm = get_tqdm()

logger = logging.getLogger('stanza.constituency.trainer')

def parse_args(args=None):
    parser = argparse.ArgumentParser(description="Script that uses multiple ensembles to find trees where both ensembles agree")

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument('--tokenized_file', type=str, default=None, help='Input file of tokenized text for parsing with parse_text.')
    input_group.add_argument('--tree_file', type=str, default=None, help='Input file of already parsed text for reparsing with parse_text.')
    parser.add_argument('--output_file', type=str, default=None, help='Where to put the output file')

    parser.add_argument('--charlm_forward_file', type=str, default=None, help="Exact path to use for forward charlm")
    parser.add_argument('--charlm_backward_file', type=str, default=None, help="Exact path to use for backward charlm")
    parser.add_argument('--wordvec_pretrain_file', type=str, default=None, help='Exact name of the pretrain file to read')

    utils.add_device_args(parser)

    parser.add_argument('--lang', default='en', help='Language to use')

    parser.add_argument('--eval_batch_size', type=int, default=50, help='How many trees to batch when running eval')
    parser.add_argument('--e1', type=str, nargs='+', default=None, help="Which model(s) to load in the first ensemble")
    parser.add_argument('--e2', type=str, nargs='+', default=None, help="Which model(s) to load in the second ensemble")

    parser.add_argument('--mode', default='predict', choices=['parse_text', 'predict'])

    # another option would be to include the tree idx in each entry in an existing saved file
    # the processing could then pick up at exactly the last known idx
    parser.add_argument('--start_tree', type=int, default=0, help='Where to start... most useful if the previous incarnation crashed')
    parser.add_argument('--end_tree', type=int, default=None, help='Where to end.  If unset, will process to the end of the file')

    retagging.add_retag_args(parser)

    args = vars(parser.parse_args())

    retagging.postprocess_args(args)
    args['num_generate'] = 0

    return args

def main():
    args = parse_args()
    utils.log_training_args(args, logger, name="ensemble")

    retag_pipeline = retagging.build_retag_pipeline(args)
    foundation_cache = retag_pipeline[0].foundation_cache if retag_pipeline else FoundationCache()

    logger.info("Building ensemble #1 out of %s", args['e1'])
    e1 = Ensemble(args, filenames=args['e1'], foundation_cache=foundation_cache)
    e1.to(args.get('device', None))
    logger.info("Building ensemble #2 out of %s", args['e2'])
    e2 = Ensemble(args, filenames=args['e2'], foundation_cache=foundation_cache)
    e2.to(args.get('device', None))

    if args['tokenized_file']:
        tokenized_sentences, _ = text_processing.read_tokenized_file(args['tokenized_file'])
    elif args['tree_file']:
        treebank = tree_reader.read_treebank(args['tree_file'])
        tokenized_sentences = [x.leaf_labels() for x in treebank]
        if args['lang'] == 'vi':
            tokenized_sentences = [[x.replace("_", " ") for x in sentence] for sentence in tokenized_sentences]
    logger.info("Read %d tokenized sentences", len(tokenized_sentences))

    all_models = e1.models + e2.models

    chunk_size = 1000
    with open(args['output_file'], 'w', encoding='utf-8') as fout:
        end_tree = len(tokenized_sentences) if args['end_tree'] is None else args['end_tree']
        for chunk_start in tqdm(range(args['start_tree'], end_tree, chunk_size)):
            chunk = tokenized_sentences[chunk_start:chunk_start+chunk_size]
            logger.info("Processing trees %d to %d", chunk_start, chunk_start+len(chunk))
            parsed1 = text_processing.parse_tokenized_sentences(args, e1, retag_pipeline, chunk)
            parsed1 = [x.predictions[0].tree for x in parsed1]
            parsed2 = text_processing.parse_tokenized_sentences(args, e2, retag_pipeline, chunk)
            parsed2 = [x.predictions[0].tree for x in parsed2]
            matching = [t for t, t2 in zip(parsed1, parsed2) if t == t2]
            logger.info("%d trees matched", len(matching))
            model_counts = [0] * len(matching)
            for model in all_models:
                model_chunk = model.parse_sentences_no_grad(iter(matching), model.build_batch_from_trees, args['eval_batch_size'], model.predict)
                model_chunk = [x.predictions[0].tree for x in model_chunk]
                for idx, (t1, t2) in enumerate(zip(matching, model_chunk)):
                    if t1 == t2:
                        model_counts[idx] += 1
            for count, tree in zip(model_counts, matching):
                line = {"tree": "%s" % tree, "count": count}
                fout.write(json.dumps(line))
                fout.write("\n")


if __name__ == '__main__':
    main()



"""
After running build_silver_dataset.py, this extracts the trees of a certain match level

For example

python3 stanza/utils/datasets/constituency/extract_silver_dataset.py --parsed_trees /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/a*.trees --keep_score 0 --output_file /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/it_silver_0.mrg

for i in `echo 0 1 2 3 4 5 6 7 8 9 10`; do python3 stanza/utils/datasets/constituency/extract_silver_dataset.py --parsed_trees /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/a*.trees --keep_score $i --output_file /u/nlp/data/constituency-parser/italian/2024_it_vit_electra/it_silver_$i.mrg; done
"""

import argparse
import json

def parse_args():
    parser = argparse.ArgumentParser(description="After finding common trees using build_silver_dataset, this extracts them all or just the ones from a particular level of accuracy")
    parser.add_argument('--parsed_trees', type=str, nargs='+', help='Input file(s) of trees parsed into the build_silver_dataset json format.')
    parser.add_argument('--keep_score', type=int, default=None, help='Which agreement level to keep.  None keeps all') 
    parser.add_argument('--output_file', type=str, default=None, help='Where to put the output file')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    trees = []
    for filename in args.parsed_trees:
        with open(filename, encoding='utf-8') as fin:
            for line in fin.readlines():
                tree = json.loads(line)
                if args.keep_score is None or tree['count'] == args.keep_score:
                    tree = tree['tree']
                    trees.append(tree)

    if args.output_file is None:
        for tree in trees:
            print(tree)
    else:
        with open(args.output_file, 'w', encoding='utf-8') as fout:
            for tree in trees:
                fout.write(tree)
                fout.write('\n')

if __name__ == '__main__':
    main()




import sys
import logging
import torch
import torch.nn as nn
import torch.optim as optim

from stanza.models.common import utils
from stanza.models.common.trainer import Trainer as BaseTrainer
from stanza.models.tokenization.utils import create_dictionary

from .model import Tokenizer
from .vocab import Vocab

logger = logging.getLogger('stanza')

class Trainer(BaseTrainer):
    def __init__(self, args=None, vocab=None, lexicon=None, dictionary=None, model_file=None, device=None, foundation_cache=None):
        # TODO: make a test of the training w/ and w/o charlm
        if model_file is not None:
            # load everything from file
            self.load(model_file, args, foundation_cache)
        else:
            # build model from scratch
            self.args = args
            self.vocab = vocab
            self.lexicon = list(lexicon) if lexicon is not None else None
            self.dictionary = dictionary
            self.model = Tokenizer(self.args, self.args['vocab_size'], self.args['emb_dim'], self.args['hidden_dim'], dropout=self.args['dropout'], feat_dropout=self.args['feat_dropout'])
        self.model = self.model.to(device)
        self.criterion = nn.CrossEntropyLoss(ignore_index=-1).to(device)
        self.optimizer = utils.get_optimizer("adam", self.model, lr=self.args['lr0'], betas=(.9, .9), weight_decay=self.args['weight_decay'])
        self.feat_funcs = self.args.get('feat_funcs', None)
        self.lang = self.args['lang'] # language determines how token normalization is done

    def update(self, inputs):
        self.model.train()
        units, labels, features, text = inputs
        lengths = [len(x) for x in text]

        device = next(self.model.parameters()).device
        units = units.to(device)
        labels = labels.to(device)
        features = features.to(device)

        pred = self.model(units, features, lengths, text)

        self.optimizer.zero_grad()
        classes = pred.size(2)
        loss = self.criterion(pred.view(-1, classes), labels.view(-1))

        loss.backward()
        nn.utils.clip_grad_norm_(self.model.parameters(), self.args['max_grad_norm'])
        self.optimizer.step()

        return loss.item()

    def predict(self, inputs):
        self.model.eval()
        units, _, features, text = inputs
        lengths = [len(x) for x in text]

        device = next(self.model.parameters()).device
        units = units.to(device)
        features = features.to(device)

        pred = self.model(units, features, lengths, text)

        return pred.data.cpu().numpy()

    def save(self, filename, skip_modules=True):
        model_state = None
        if self.model is not None:
            model_state = self.model.state_dict()
            # skip saving modules like the pretrained charlm
            if skip_modules:
                skipped = [k for k in model_state.keys() if k.split('.')[0] in self.model.unsaved_modules]
                for k in skipped:
                    del model_state[k]

        params = {
            'model': model_state,
            'vocab': self.vocab.state_dict(),
            # save and load lexicon as list instead of set so
            # we can use weights_only=True
            'lexicon': list(self.lexicon) if self.lexicon is not None else None,
            'config': self.args
        }
        try:
            torch.save(params, filename, _use_new_zipfile_serialization=False)
            logger.info("Model saved to {}".format(filename))
        except BaseException:
            logger.warning("Saving failed... continuing anyway.")

    def load(self, filename, args, foundation_cache):
        try:
            checkpoint = torch.load(filename, lambda storage, loc: storage, weights_only=True)
        except BaseException:
            logger.error("Cannot load model from {}".format(filename))
            raise
        self.args = checkpoint['config']
        if args is not None and args.get('charlm_forward_file', None) is not None:
            if checkpoint['config'].get('charlm_forward_file') is None:
                # if the saved model didn't use a charlm, we skip the charlm here
                # otherwise the loaded model weights won't fit in the newly created model
                self.args['charlm_forward_file'] = None
            else:
                self.args['charlm_forward_file'] = args['charlm_forward_file']
        if self.args.get('use_mwt', None) is None:
            # Default to True as many currently saved models
            # were built with mwt layers
            self.args['use_mwt'] = True
        self.model = Tokenizer(self.args, self.args['vocab_size'], self.args['emb_dim'], self.args['hidden_dim'], dropout=self.args['dropout'], feat_dropout=self.args['feat_dropout'])
        self.model.load_state_dict(checkpoint['model'], strict=False)
        self.vocab = Vocab.load_state_dict(checkpoint['vocab'])
        self.lexicon = checkpoint['lexicon']

        if self.lexicon is not None:
            self.lexicon = set(self.lexicon)
            self.dictionary = create_dictionary(self.lexicon)
        else:
            self.dictionary = None



from collections import namedtuple
import glob
import os
import sys
import xml.etree.ElementTree as ET

from stanza.models.classifiers.data import SentimentDatum
import stanza.utils.datasets.sentiment.process_utils as process_utils

ArguanaSentimentDatum = namedtuple('ArguanaSentimentDatum', ['begin', 'end', 'rating'])

"""
Extracts positive, neutral, and negative phrases from the ArguAna hotel review corpus

Run as follows:

python3 parse_arguana_xml.py split/training data/sentiment

ArguAna can be downloaded here:

http://argumentation.bplaced.net/arguana/data
http://argumentation.bplaced.net/arguana-data/arguana-tripadvisor-annotated-v2.zip
"""

def get_phrases(filename):
    tree = ET.parse(filename)
    fragments = []

    root = tree.getroot()
    body = None
    for child in root:
        if child.tag == '{http:///uima/cas.ecore}Sofa':
            body = child.attrib['sofaString']
        elif child.tag == '{http:///de/aitools/ie/uima/type/arguana.ecore}Fact':
            fragments.append(ArguanaSentimentDatum(begin=int(child.attrib['begin']),
                                                   end=int(child.attrib['end']),
                                                   rating="1"))
        elif child.tag == '{http:///de/aitools/ie/uima/type/arguana.ecore}Opinion':
            if child.attrib['polarity'] == 'negative':
                rating = "0"
            elif child.attrib['polarity'] == 'positive':
                rating = "2"
            else:
                raise ValueError("Unexpected polarity found in {}".format(filename))
            fragments.append(ArguanaSentimentDatum(begin=int(child.attrib['begin']),
                                                   end=int(child.attrib['end']),
                                                   rating=rating))


    phrases = [SentimentDatum(fragment.rating, body[fragment.begin:fragment.end]) for fragment in fragments]
    #phrases = [phrase.replace("\n", " ") for phrase in phrases]
    return phrases

def get_phrases_from_directory(directory):
    phrases = []
    inpath = os.path.join(directory, "arguana-tripadvisor-annotated-v2", "split", "training", "*", "*xmi")
    for filename in glob.glob(inpath):
        phrases.extend(get_phrases(filename))
    return phrases

def get_tokenized_phrases(in_directory):
    phrases = get_phrases_from_directory(in_directory)
    phrases = process_utils.get_ptb_tokenized_phrases(phrases)
    print("Found {} phrases in arguana".format(len(phrases)))
    return phrases

def main(in_directory, out_directory, short_name):
    phrases = get_tokenized_phrases(in_directory)
    process_utils.write_list(os.path.join(out_directory, "%s.train.json" % short_name), phrases)


if __name__ == "__main__":
    in_directory = sys.argv[1]
    out_directory = sys.argv[2]
    short_name = sys.argv[3]
    main(in_directory, out_directory, short_name)



import logging

import torch
from torch import nn
import torch.nn.functional as F

logger = logging.getLogger('stanza')

class RelativeAttention(nn.Module):
    def __init__(self, d_model, num_heads, window=8, dropout=0.2, reverse=False, d_output=None, fudge_output=False, num_sinks=0):
        super().__init__()
        if d_output is None:
            d_output = d_model

        d_head, remainder = divmod(d_output, num_heads)
        if remainder:
            if fudge_output:
                d_head = d_head + 1
                logger.debug("Relative attn: %d %% %d != 0, updating d_output to %d", d_output, num_heads, num_heads * d_head)
                d_output = num_heads * d_head
            else:
                raise ValueError("incompatible `d_model` and `num_heads`")
        self.window = window
        self.num_sinks = num_sinks
        self.d_model = d_model
        self.d_head = d_head
        self.num_heads = num_heads
        self.d_output = d_output
        self.key = nn.Linear(d_model, d_output)
        # the bias for query all gets trained to 0 anyway
        self.query = nn.Linear(d_model, d_output, bias=False)
        self.value = nn.Linear(d_model, d_output, bias=False)
        # initializing value with eye seems to hurt!
        #nn.init.eye_(self.value.weight)

        self.dropout = nn.Dropout(dropout)
        self.position = nn.Parameter(torch.randn(1, 1, d_head, window + num_sinks, 1))

        self.register_buffer(
            "mask", 
            torch.tril(torch.ones(window, window), diagonal=-1).unsqueeze(0).unsqueeze(0).unsqueeze(0)
        )
        self.register_buffer(
            "flipped_mask",
            torch.flip(self.mask, (-1,))
        )

        self.reverse = reverse

    def forward(self, x, sink=None):
        # x.shape == (batch_size, seq_len, d_model)
        batch_size, seq_len, d_model = x.shape
        if d_model != self.d_model:
            raise ValueError("Incompatible input")

        if self.reverse:
            x = torch.flip(x, (1,))

        orig_seq_len = seq_len
        if seq_len < self.window:
            zeros = torch.zeros((x.shape[0], self.window - seq_len, x.shape[2]), dtype=x.dtype, device=x.device)
            x = torch.cat((x, zeros), axis=1)
            seq_len = self.window

        if self.num_sinks > 0:
            # could keep a parameter to train sinks, but as it turns out,
            # the position vectors just overlap that parameter space anyway
            # generally the model trains the sinks to zero if we do that
            if sink is None:
                sink = torch.zeros((batch_size, self.num_sinks, d_model), dtype=x.dtype, device=x.device)
            else:
                sink = sink.expand(batch_size, self.num_sinks, d_model)
            x = torch.cat((sink, x), axis=1)

        # k.shape = (batch_size, num_heads, d_head, seq_len + num_sinks)
        k = self.key(x).reshape(batch_size, seq_len + self.num_sinks, self.num_heads, -1).permute(0, 2, 3, 1)[:, :, :, self.num_sinks:]

        # v.shape = (batch_size, num_heads, d_head, seq_len + num_sinks)
        v = self.value(x).reshape(batch_size, seq_len + self.num_sinks, self.num_heads, -1).permute(0, 2, 3, 1)

        # q.shape = (batch_size, num_heads, d_head, seq_len + num_sinks)
        q = self.query(x).reshape(batch_size, seq_len + self.num_sinks, self.num_heads, -1).permute(0, 2, 3, 1)
        # q.shape = (batch_size, num_heads, d_head, window + num_sinks, seq_len)
        q = self.skew_repeat(q)
        q = q + self.position

        # qk.shape = (batch_size, num_heads, d_head, window + num_sinks, seq_len)
        qk = torch.einsum('bndws,bnds->bndws', q, k)

        # TODO: fix mask
        # mask out the padding spaces at the end
        # can only attend to spots that aren't padded
        if orig_seq_len < seq_len:
            # mask out the part of the sentence which is empty
            shorter_mask = self.flipped_mask[:, :, :, :orig_seq_len, -orig_seq_len:]
            qk = qk[:, :, :, :(orig_seq_len + self.num_sinks), :orig_seq_len]
            qk[:, :, :, -orig_seq_len:, :] = qk[:, :, :, -orig_seq_len:, :].masked_fill(shorter_mask == 1, float("-inf"))
        else:
            qk[:, :, :, -self.window:, -self.window:] = qk[:, :, :, -self.window:, -self.window:].masked_fill(self.flipped_mask == 1, float("-inf"))
        qk = F.softmax(qk, dim=3)

        # v.shape = (batch_size, num_heads, d_head, window, seq_len)
        v = self.skew_repeat(v)
        if orig_seq_len < seq_len:
            v = v[:, :, :, :(orig_seq_len + self.num_sinks), :orig_seq_len]
        # result.shape = (batch_size, num_heads, d_head, orig_seq_len)
        result = torch.einsum('bndws,bndws->bnds', qk, v)
        # batch_size, orig_seq_len, d_output
        result = result.reshape(batch_size, self.d_output, orig_seq_len).transpose(1, 2)

        if self.reverse:
            result = torch.flip(result, (1,))

        return self.dropout(result)

    def skew_repeat(self, q):
        """
        q (currently, at least) is num_sinks + seq_len long
        and the num_sinks are there to be chopped off the front
        then the seq_len remainder is skewed
        """
        if self.num_sinks > 0:
            q_sink = q[:, :, :, :self.num_sinks]
            q_sink = q_sink.unsqueeze(4)
            q_sink = q_sink.repeat(1, 1, 1, 1, q.shape[-1] - self.num_sinks)
            q = q[:, :, :, self.num_sinks:]
        # make stripes that look like this
        # (seq_len 5, window 3)
        #   1 2 3 4 5
        #   1 2 3 4 5
        #   1 2 3 4 5
        q = q.unsqueeze(4).repeat(1, 1, 1, 1, self.window).transpose(3, 4)
        # now the stripes look like
        #   1 2 3 4 5
        #   0 2 3 4 5
        #   0 0 3 4 5
        q[:, :, :, :, :self.window] = q[:, :, :, :, :self.window].masked_fill(self.mask == 1, 0)
        q_shape = list(q.shape)
        q_new_shape = list(q.shape)[:-2] + [-1]
        q = q.reshape(q_new_shape)
        zeros = torch.zeros_like(q[:, :, :, :1])
        zeros = zeros.repeat(1, 1, 1, self.window)
        q = torch.cat((q, zeros), axis=-1)
        q_new_shape = q_new_shape[:-1] + [self.window, -1]
        # now the stripes look like
        #   1 2 3 4 5
        #   2 3 4 5 0
        #   3 4 5 0 0
        # q.shape = (batch_size, num_heads, d_head, window, seq_len)
        q = q.reshape(q_new_shape)[:, :, :, :, :-1]
        if self.num_sinks > 0:
            q = torch.cat([q_sink, q], dim=3)
        return q



import argparse
import sys

import stanza.utils.datasets.common as common

def remove_column(sentences, column):
    new_sentences = []
    for sentence in sentences:
        new_lines = []
        for line in sentence:
            if line.startswith("#"):
                new_lines.append(line)
            else:
                pieces = line.split("\t")
                pieces[column] = "_"
                new_lines.append("\t".join(pieces))
        new_sentences.append(new_lines)
    return new_sentences

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--infile', help='Which file to manipulate')
    parser.add_argument('--outfile', help='Where to write the result')
    args = parser.parse_args()

    sentences = common.read_sentences_from_conllu(args.infile)
    sentences = [common.maybe_add_fake_dependencies(sentence) for sentence in sentences]
    sentences = remove_column(sentences, 4)
    sentences = remove_column(sentences, 5)
    common.write_sentences_to_conllu(args.outfile, sentences)
    
if __name__ == "__main__":
    main()



"""
Processor for performing lemmatization
"""

from itertools import compress

import torch

from stanza.models.common import doc
from stanza.models.lemma.data import DataLoader
from stanza.models.lemma.trainer import Trainer
from stanza.pipeline._constants import *
from stanza.pipeline.processor import UDProcessor, register_processor

WORD_TAGS = [doc.TEXT, doc.UPOS]

@register_processor(name=LEMMA)
class LemmaProcessor(UDProcessor):

    # set of processor requirements this processor fulfills
    PROVIDES_DEFAULT = set([LEMMA])
    # set of processor requirements for this processor
    # pos will be added later for non-identity lemmatizerx
    REQUIRES_DEFAULT = set([TOKENIZE])
    # default batch size
    DEFAULT_BATCH_SIZE = 5000

    def __init__(self, config, pipeline, device):
        # run lemmatizer in identity mode
        self._use_identity = None
        self._pretagged = None
        super().__init__(config, pipeline, device)

    @property
    def use_identity(self):
        return self._use_identity

    def _set_up_model(self, config, pipeline, device):
        if config.get('use_identity') in ['True', True]:
            self._use_identity = True
            self._config = config
            self.config['batch_size'] = LemmaProcessor.DEFAULT_BATCH_SIZE
        else:
            # the lemmatizer only looks at one word when making
            # decisions, not the surrounding context
            # therefore, we can save some time by remembering what
            # we did the last time we saw any given word,pos
            # since a long running program will remember everything
            # (unless we go back and make it smarter)
            # we make this an option, not the default
            # TODO: need to update the cache to skip the contextual lemmatizer
            self.store_results = config.get('store_results', False)
            self._use_identity = False
            args = {'charlm_forward_file': config.get('forward_charlm_path', None),
                    'charlm_backward_file': config.get('backward_charlm_path', None)}
            lemma_classifier_args = dict(args)
            lemma_classifier_args['wordvec_pretrain_file'] = config.get('pretrain_path', None)
            self._trainer = Trainer(args=args, model_file=config['model_path'], device=device, foundation_cache=pipeline.foundation_cache, lemma_classifier_args=lemma_classifier_args)

    def _set_up_requires(self):
        self._pretagged = self._config.get('pretagged', None)
        if self._pretagged:
            self._requires = set()
        elif self.config.get('pos') and not self.use_identity:
            self._requires = LemmaProcessor.REQUIRES_DEFAULT.union(set([POS]))
        else:
            self._requires = LemmaProcessor.REQUIRES_DEFAULT

    def process(self, document):
        if not self.use_identity:
            batch = DataLoader(document, self.config['batch_size'], self.config, vocab=self.vocab, evaluation=True, expand_unk_vocab=True)
        else:
            batch = DataLoader(document, self.config['batch_size'], self.config, evaluation=True, conll_only=True)
        if self.use_identity:
            preds = [word.text for sent in batch.doc.sentences for word in sent.words]
        elif self.config.get('dict_only', False):
            preds = self.trainer.predict_dict(batch.doc.get([doc.TEXT, doc.UPOS]))
        else:
            if self.config.get('ensemble_dict', False):
                # skip the seq2seq model when we can
                skip = self.trainer.skip_seq2seq(batch.doc.get([doc.TEXT, doc.UPOS]))
                # although there is no explicit use of caseless or lemma_caseless in this processor,
                # it shows up in the config which gets passed to the DataLoader,
                # possibly affecting its results
                seq2seq_batch = DataLoader(document, self.config['batch_size'], self.config, vocab=self.vocab,
                                           evaluation=True, skip=skip, expand_unk_vocab=True)
            else:
                seq2seq_batch = batch

            with torch.no_grad():
                preds = []
                edits = []
                for i, b in enumerate(seq2seq_batch):
                    ps, es = self.trainer.predict(b, self.config['beam_size'], seq2seq_batch.vocab)
                    preds += ps
                    if es is not None:
                        edits += es

            if self.config.get('ensemble_dict', False):
                word_tags = batch.doc.get(WORD_TAGS)
                words = [x[0] for x in word_tags]
                preds = self.trainer.postprocess([x for x, y in zip(words, skip) if not y], preds, edits=edits)
                if self.store_results:
                    new_word_tags = compress(word_tags, map(lambda x: not x, skip))
                    new_predictions = [(x[0], x[1], y) for x, y in zip(new_word_tags, preds)]
                    self.trainer.train_dict(new_predictions, update_word_dict=False)
                # expand seq2seq predictions to the same size as all words
                i = 0
                preds1 = []
                for s in skip:
                    if s:
                        preds1.append('')
                    else:
                        preds1.append(preds[i])
                        i += 1
                preds = self.trainer.ensemble(word_tags, preds1)
            else:
                preds = self.trainer.postprocess(batch.doc.get([doc.TEXT]), preds, edits=edits)

            if self.trainer.has_contextual_lemmatizers():
                preds = self.trainer.update_contextual_preds(batch.doc, preds)

        # map empty string lemmas to '_'
        preds = [max([(len(x), x), (0, '_')])[1] for x in preds]
        batch.doc.set([doc.LEMMA], preds)
        return batch.doc



""" Contains type aliases for coref module """

from dataclasses import dataclass
from typing import Any, Dict, List, Tuple

import torch


EPSILON = 1e-7
LARGE_VALUE = 1000  # used instead of inf due to bug #16762 in pytorch

Doc = Dict[str, Any]
Span = Tuple[int, int]


@dataclass
class CorefResult:
    coref_scores: torch.Tensor = None                  # [n_words, k + 1]
    coref_y: torch.Tensor = None                       # [n_words, k + 1]
    rough_y: torch.Tensor = None                       # [n_words, n_words]

    word_clusters: List[List[int]] = None
    span_clusters: List[List[Span]] = None

    rough_scores: torch.Tensor = None                  # [n_words, n_words]
    span_scores: torch.Tensor = None                   # [n_heads, n_words, 2]
    span_y: Tuple[torch.Tensor, torch.Tensor] = None   # [n_heads] x2

    zero_scores: torch.Tensor = None



import argparse
import json
import os
import random
import tarfile
import tempfile
from tqdm import tqdm
# could import lxml here, but that would involve adding lxml as a
# dependency to the stanza package
# another alternative would be to try & catch ImportError
try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree


NAMESPACE = "http://www.tei-c.org/ns/1.0"
MORPH_FILE = "ann_morphosyntax.xml"
NER_FILE = "ann_named.xml"
SEGMENTATION_FILE = "ann_segmentation.xml"

def parse_xml(path):
    if not os.path.exists(path):
        return None
    et = etree.parse(path)
    rt = et.getroot()
    return rt


def get_node_id(node):
    # get the id from the xml node
    return node.get('{http://www.w3.org/XML/1998/namespace}id')


def extract_entities_from_subfolder(subfolder, nkjp_dir):
    # read the ner annotation from a subfolder, assign it to paragraphs
    subfolder_entities = extract_unassigned_subfolder_entities(subfolder, nkjp_dir)
    par_id_to_segs = assign_entities(subfolder, subfolder_entities, nkjp_dir)
    return par_id_to_segs


def extract_unassigned_subfolder_entities(subfolder, nkjp_dir):
    """
    Build and return a map from par_id to extracted entities
    """
    ner_path = os.path.join(nkjp_dir, subfolder, NER_FILE)
    rt = parse_xml(ner_path)
    if rt is None:
        return None
    subfolder_entities = {}
    ner_pars = rt.findall("{%s}TEI/{%s}text/{%s}body/{%s}p" % (NAMESPACE, NAMESPACE, NAMESPACE, NAMESPACE))
    for par in ner_pars:
        par_entities = {}
        _, par_id = get_node_id(par).split("_")
        ner_sents = par.findall("{%s}s" % NAMESPACE)
        for ner_sent in ner_sents:
            corresp = ner_sent.get("corresp")
            _, ner_sent_id  = corresp.split("#morph_")
            par_entities[ner_sent_id] = extract_entities_from_sentence(ner_sent)
        subfolder_entities[par_id] = par_entities
    return subfolder_entities

def extract_entities_from_sentence(ner_sent):
    # extracts all the entity dicts from the sentence
    # we assume that an entity cannot span across sentences
    segs = ner_sent.findall("./{%s}seg" % NAMESPACE)
    sent_entities = {}
    for i, seg in enumerate(segs):
        ent_id = get_node_id(seg)
        targets = [ptr.get("target") for ptr in seg.findall("./{%s}ptr" % NAMESPACE)]
        orth = seg.findall("./{%s}fs/{%s}f[@name='orth']/{%s}string" % (NAMESPACE, NAMESPACE, NAMESPACE))[0].text
        ner_type = seg.findall("./{%s}fs/{%s}f[@name='type']/{%s}symbol" % (NAMESPACE, NAMESPACE, NAMESPACE))[0].get("value")
        ner_subtype_node = seg.findall("./{%s}fs/{%s}f[@name='subtype']/{%s}symbol" % (NAMESPACE, NAMESPACE, NAMESPACE))
        if ner_subtype_node:
            ner_subtype = ner_subtype_node[0].get("value")
        else:
            ner_subtype = None
        entity = {"ent_id": ent_id,
                  "index": i,
                  "orth": orth,
                  "ner_type": ner_type,
                  "ner_subtype": ner_subtype,
                  "targets": targets}
        sent_entities[ent_id] = entity
    cleared_entities = clear_entities(sent_entities)
    return cleared_entities


def clear_entities(entities):
    # eliminates entities which extend beyond our scope
    resolve_entities(entities)
    entities_list = sorted(list(entities.values()), key=lambda ent: ent["index"])
    entities = eliminate_overlapping_entities(entities_list)
    for entity in entities:
        targets = entity["targets"]
        entity["targets"] = [t.split("morph_")[1] for t in targets]
    return entities


def resolve_entities(entities):
    # assign morphological level targets to entities
    resolved_targets = {entity_id: resolve_entity(entity, entities) for entity_id, entity in entities.items()}
    for entity_id in entities:
        entities[entity_id]["targets"] = resolved_targets[entity_id]


def resolve_entity(entity, entities):
    # translate targets defined in terms of entities, into morphological units
    # works recurrently
    targets = entity["targets"]
    resolved = []
    for target in targets:
        if target.startswith("named_"):
            target_entity = entities[target]
            resolved.extend(resolve_entity(target_entity, entities))
        else:
            resolved.append(target)
    return resolved


def eliminate_overlapping_entities(entities_list):
    # we eliminate entities which are at least partially contained in one ocurring prior to them
    # this amounts to removing overlap
    subsumed = set([])
    for sub_i, sub in enumerate(entities_list):
        for over in entities_list[:sub_i]:
            if any([target in over["targets"] for target in sub["targets"]]):
                subsumed.add(sub["ent_id"])
    return [entity for entity in entities_list if entity["ent_id"] not in subsumed]


def assign_entities(subfolder, subfolder_entities, nkjp_dir):
    # recovers all the segments from a subfolder, and annotates it with NER
    morph_path = os.path.join(nkjp_dir, subfolder, MORPH_FILE)
    rt = parse_xml(morph_path)
    morph_pars = rt.findall("{%s}TEI/{%s}text/{%s}body/{%s}p" % (NAMESPACE, NAMESPACE, NAMESPACE, NAMESPACE))
    par_id_to_segs = {}
    for par in morph_pars:
        _, par_id = get_node_id(par).split("_")
        morph_sents = par.findall("{%s}s" % NAMESPACE)
        sent_id_to_segs = {}
        for morph_sent in morph_sents:
            _, sent_id = get_node_id(morph_sent).split("_")
            segs = morph_sent.findall("{%s}seg" % NAMESPACE)
            sent_segs = {}
            for i, seg in enumerate(segs):
                _, seg_id = get_node_id(seg).split("morph_")
                orth = seg.findall("{%s}fs/{%s}f[@name='orth']/{%s}string" % (NAMESPACE, NAMESPACE, NAMESPACE))[0].text
                token = {"seg_id": seg_id,
                          "i": i,
                          "orth": orth,
                          "text": orth,
                          "tag": "_",
                          "ner": "O", # This will be overwritten
                          "ner_subtype": None,
                          }
                sent_segs[seg_id] = token
            sent_id_to_segs[sent_id] = sent_segs
        par_id_to_segs[par_id] = sent_id_to_segs

    if subfolder_entities is None:
        return None

    for par_key in subfolder_entities:
        par_ents = subfolder_entities[par_key]
        for sent_key in par_ents:
            sent_entities = par_ents[sent_key]
            for entity in sent_entities:
                targets = entity["targets"]
                iob = "B"
                ner_label = entity["ner_type"]
                matching_tokens = sorted([par_id_to_segs[par_key][sent_key][target] for target in targets], key=lambda x:x["i"])
                for token in matching_tokens:
                    full_label = f"{iob}-{ner_label}"
                    token["ner"] = full_label
                    token["ner_subtype"] = entity["ner_subtype"]
                    iob = "I"
    return par_id_to_segs


def load_xml_nkjp(nkjp_dir):
    subfolder_to_annotations = {}
    subfolders = sorted(os.listdir(nkjp_dir))
    for subfolder in tqdm([name for name in subfolders if os.path.isdir(os.path.join(nkjp_dir, name))]):
        out = extract_entities_from_subfolder(subfolder, nkjp_dir)
        if out:
            subfolder_to_annotations[subfolder] = out
        else:
            print(subfolder, "has no ann_named.xml file")

    return subfolder_to_annotations


def split_dataset(dataset, shuffle=True, train_fraction=0.9, dev_fraction=0.05, test_section=True):
    random.seed(987654321)
    if shuffle:
        random.shuffle(dataset)

    if not test_section:
        dev_fraction = 1 - train_fraction

    train_size = int(train_fraction * len(dataset))
    dev_size = int(dev_fraction * len(dataset))
    train = dataset[:train_size]
    dev = dataset[train_size: train_size + dev_size]
    test = dataset[train_size + dev_size:]

    return {
        'train': train,
        'dev': dev,
        'test': test
    }


def convert_nkjp(nkjp_path, output_dir):
    """Converts NKJP NER data into IOB json format.

    nkjp_dir is the path to directory where NKJP files are located.
    """
    # Load XML NKJP
    print("Reading data from %s" % nkjp_path)
    if os.path.isfile(nkjp_path) and (nkjp_path.endswith(".tar.gz") or nkjp_path.endswith(".tgz")):
        with tempfile.TemporaryDirectory() as nkjp_dir:
            print("Temporarily extracting %s to %s" % (nkjp_path, nkjp_dir))
            with tarfile.open(nkjp_path, "r:gz") as tar_in:
                tar_in.extractall(nkjp_dir)

            subfolder_to_entities = load_xml_nkjp(nkjp_dir)
    elif os.path.isdir(nkjp_path):
        subfolder_to_entities = load_xml_nkjp(nkjp_path)
    else:
        raise FileNotFoundError("Cannot find either unpacked dataset or gzipped file")
    converted = []
    for subfolder_name, pars in subfolder_to_entities.items():
        for par_id, par in pars.items():
            paragraph_identifier = f"{subfolder_name}|{par_id}"
            par_tokens = []
            for _, sent in par.items():
                tokens = sent.values()
                srt = sorted(tokens, key=lambda tok:tok["i"])
                for token in srt:
                    _ = token.pop("i")
                    _ = token.pop("seg_id")
                    par_tokens.append(token)
            par_tokens[0]["paragraph_id"] = paragraph_identifier
            converted.append(par_tokens)

    split = split_dataset(converted)

    for split_name, split in split.items():
        if split:
            with open(os.path.join(output_dir, f"pl_nkjp.{split_name}.json"), "w", encoding="utf-8") as f:
                json.dump(split, f, ensure_ascii=False, indent=2)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input_path', type=str, default="/u/nlp/data/ner/stanza/polish/NKJP-PodkorpusMilionowy-1.2.tar.gz", help="Where to find the files")
    parser.add_argument('--output_path', type=str, default="data/ner", help="Where to output the results")
    args = parser.parse_args()

    convert_nkjp(args.input_path, args.output_path)


if __name__ == '__main__':
    main()
